#ifndef _BIBLIN_H_
#define _BIBLIN_H_
/*
 * biblin.h
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Contains:
 *    A header file for Bit Buffer module (see module description in biblin.c)
 */

#include "nrctypes.h"

#ifdef BIB_VDEBUF
#include "dlist.h"
#ifdef BIB_H223_BUF
#include "h223buf.h"
#endif
#endif


/*
 * Defines
 */

#define ERR_BIB_STRUCT_ALLOC 1000
#define ERR_BIB_BUFFER_ALLOC 1001
#define ERR_BIB_FILE_READ 1002
#define ERR_BIB_NOT_ENOUGH_DATA 1003
#define ERR_BIB_ALREADY_OPENED 1004
#define ERR_BIB_FILE_OPEN 1005
#define ERR_BIB_ALREADY_CLOSED 1006
#define ERR_BIB_FILE_CLOSE 1007
#define ERR_BIB_NUM_BITS 1008
#define ERR_BIB_FILE_NOT_OPEN 1009
#define ERR_BIB_ILLEGAL_SIZE 1010
#define ERR_BIB_CANNOT_REWIND 1011
#define ERR_BIB_BUFLIST 1012
#define ERR_BIB_TOO_SMALL_BUFFER 1013
#define ERR_BIB_FEC_RELOCK 1050


/*
 * Structs and typedefs
 */

#ifdef BIB_VDEBUF
typedef struct {
   DLST_ITEM_SKELETON

   void *buf;                 /* a pointer to a buffer structure */
} bibListItem_t;
#endif

/* {{-output"bibbufti.txt" -ignore"*" -noCR}}
 * The bibBuffer_t data type is a structure containing all the necessary data
 * for a bit buffer instance (except for the actual data buffer). This
 * structure is passed to all of the bit buffer functions.
 * {{-output"bibbufti.txt"}}
 */

/* {{-output"bibbuft.txt"}} */
typedef struct bibBuffer_s {
   u_char *baseAddr;       /* the start address of the buffer */
   unsigned size;          /* the size of the buffer in bytes */
   unsigned getIndex;      /* an index to the buffer where data was last got */
   int bitIndex;           /* an index to the byte pointed by getIndex + 1 */
   u_int32 bitsLeft;       /* the number of bits currently in the buffer */
   u_int32 numBytesRead;   /* the total number of bytes read */
#ifdef BIB_VDEBUF
   dlst_t bufList;         /* list of input buffers */
   u_int32 numBitsBeforeCurrListItem;
#endif
#ifdef BIB_PSC_INSERT
   int numEmulPreventBits;
#endif
} bibBuffer_t;
/* {{-output"bibbuft.txt"}} */



/*
 * External macros
 */

#ifdef BIB_PSC_INSERT
   #define bibNumberOfEmulationPreventionBits(buffer) \
      ((buffer)->numEmulPreventBits)
   #define bibNumberOfBitsLeft(buffer) \
      ((buffer)->bitsLeft)
#endif


/*
 * External function prototypes
 */

bibBuffer_t *bibCreate(void *srcBuffer, unsigned srcBufferLength,
   int16 *errorCode);

void bibDelete(bibBuffer_t **buffer, int16 *errorCode);

u_int32 bibNumberOfFlushedBits(bibBuffer_t *buffer);

u_int32 bibNumberOfFlushedBytes(bibBuffer_t *buffer);

u_int32 bibNumberOfRewBits(bibBuffer_t *buffer);

void bibRewindBits(u_int32 numberOfBits, bibBuffer_t *buffer, int16 *errorCode);

/* 
 * Prototypes for bibFlushBits/bibGetBits/bibShowBits 
 */

/* If end of buffer is not reported (by ERR_BIB_NOT_ENOUGH_DATA) */
#if defined(BIB_NO_EOB)

   /* These prototypes contain an extra parameter telling whether the read
      bits contain a part of a synchronization code */
   void bibFlushBits(int numberOfBits, bibBuffer_t *buffer,
      int *numberOfBitsGot, int *syncStartIndex, int16 *errorCode);

   u_int32 bibGetBits(int numberOfBits, bibBuffer_t *buffer,
      int *numberOfBitsGot, int *syncStartIndex, int16 *errorCode);

   u_int32 bibShowBits(int numberOfBits, bibBuffer_t *buffer,
      int *numberOfBitsGot, int *syncStartIndex, int16 *errorCode);

/* Else */
#else

   /* "Normal" prototypes */
   void bibFlushBits(int numberOfBits, bibBuffer_t *buffer,
      int *numberOfBitsGot, int16 *errorCode);

   u_int32 bibGetBits(int numberOfBits, bibBuffer_t *buffer,
      int *numberOfBitsGot, int16 *errorCode);

   u_int32 bibShowBits(int numberOfBits, bibBuffer_t *buffer,
      int *numberOfBitsGot, int16 *errorCode);
#endif

#endif
