#ifndef _FIL_H_
#define _FIL_H_


/*
 * deblockFilt.h
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Contains:
 *    Functions for deblocking filter.
 *
 */

#include "nrctypes.h"

/*
 * Defines
 */


/*
 * Structs and typedefs
 */


/*
 * Function prototypes
 */

/*
 *
 * filLoopFilter:
 *
 * Parameters:
 *      orig      I: Original frame
 *      picSizeX  I: Luminance frame width
 *      picSizeY  I: Luminance frame height
 *      qp        I: Quantization parameter
 *      result    O: Result of filtering
 *
 * Function:
 *      Calls filCompFilt() function, which performs loop filtering for 
 *      luminance and chrominance frames.
 *
 * Returns:
 *      Nothing
 */

void filCompFilt(u_char *orig, int32 picSizeX, int32 picSizeY,
                 int32 qp, u_char *result, int intra_flag, int enco_flag, 
                 object *object_list1, objectBuffer_t *object_list2,
                 int useLoopFilter);

void filLoopFilter(struct frame *orig, int32 picSizeX, int32 picSizeY,
                   int32 qp, struct frame *result, int useLoopFilter);

/*
 *
 * filCompFiltIntra:
 *
 * Parameters:
 *      orig      I: Original frame
 *      picSizeX  I: Luminance frame width
 *      picSizeY  I: Luminance frame height
 *      qp        I: Quantization parameter
 *      upLeftX,
 *      upLeftY   I: Corner coordinates of the 8x8 block to be filtered
 *      result    O: Result of filtering
 *
 * Function:
 *      Performs loop filtering for one 8x8 block inside the frame
 *      using average filter with varying filter support.
 *
 * Returns:
 *      Nothing
 */

void filCompFiltIntra(u_char *orig, int32 picSizeX, int32 picSizeY,
                      int32 qp, u_char *result,
                      int32 upLeftX, int32 upLeftY);


/*
 *
 * filInitDiffActTable:
 *
 * Parameters:
 *      Nothing
 *
 * Function:
 *      Initializes table for local activity parameter, defined by pixel  
 *      value difference within block and quantization parameter value.
 *
 * Returns:
 *      Nothing
 */

void filInitDiffActTable(void);


/*
 *
 * filInitOverallActTable:
 *
 * Parameters:
 *      Nothing
 *
 * Function:
 *      Initializes table for maximum search support corresponding to
 *      overall activity, defined by pixel value difference at the block 
 *      boundary and quantization parameter value. 
 *
 * Returns:
 *      Nothing
 */

void filInitOverallActTable(void);


/*
 *
 * fil_form_region_and_pec_info()
 *
 * Parameters:
 *      i1          : x coordinate parameter for the current macroblock
 *      j1          : y coordinate parameter for the current macroblock
 *      mbsize      : macroblock size parameter
 *      picsizex    : picture size x
 *      picsizey    : picture size y
 *      object_list : macroblock information list  
 *      intra_flag  : intra frame indication (if 1 => intra frame ( means no usage of object_list info))
 *      enco_flag   : encoder indication flag ( if 1 => encoder running , use object_list1 else object_list2)
 *      regmodes[4] : contanes the regionmodes of the neighboring blocks
 *      pecmodes[4] : containes the pecmodes of the neighboring blocks
 * 
 * Function:
 *      Creates the region type and pec type info of the blocks inside the macroblock sized 
 *      region being filtered.
 *      CAUTION : The fixed point operations ( like bit shifts ) assume that  Macroblock size is 16!!!
 *      
 *
 * Returns:
 *      regmodes[4] : contanes the regionmodes of the neighboring blocks
 *
 *      pecmodes[4] : containes the pecmodes of the neighboring blocks
 */

void fil_form_region_and_pec_info(int i1, int j1, int mbsize, int picsizex, int picsizey, 
                                    object *object_list1, objectBuffer_t *object_list2, 
                                    int intra_flag, int enco_flag, int *regmodes, int *pecmodes);

/*
 *
 * fil_overall_activity_truncator()
 *
 * Parameters:
 *      index       : the line being processed 
 *      regmodes[4] : contanes the regionmodes of the neighboring blocks
 *      pecmodes[4] : containes the pecmodes of the neighboring blocks
 *      rightTrunc : overall activity truncation value for the rightside of the block boundary
 *      leftTrunc  : overall activity truncation value for the leftside of the block boundary
 *      direction_flag : indicates the direction of filtering. 1 : horizontal 0 : vertical filtering
 *
 * Function:
 *      Truncates the overall activity parameter according to the regionMode and pecMode info.
 *      CAUTION     : macroblock size is assumed to be 16!!!
 *
 *
 * Returns:
 *      rightTrunc : overall activity truncation value for the rightside of the block boundary
 *      leftTrunc  : overall activity truncation value for the leftside of the block boundary
 */

void fil_overall_activity_truncator(int direction_flag, int index, int *regmodes, int *pecmodes, 
                                    int *rightTrunc, int *leftTrunc);                
#endif
