/*
 * decDct.c
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Contains:
 *    Decoding functions for reconstructing the prediction error from 
 *    the coding information.
 *
 */

#include "decDct.h"


extern void varSizeInvDctFixedp(int* src,int* dest, int sidx, int sidy);


/*
 * Define extern tables.
 */

extern u_char 
  zig8x8[64],
  zig4x4[16];

/*
 * decDct:
 *
 * Parameters:
 *	qcoeffs		I: Quantized dct coefficients pointer
 *	recoBlock	O: Reconstruction block pointer
 *	qp		I: Quantization Parameter value
 *	rows		I: Width of block
 *	cols		I: Height of block
 *
 * Function:
 *	Calculates reconstruction block for given dct coefficients and qp.
 *      The size of reconstructed block is cols times rows.
 *
 * Returns:
 *	void
 */

void decDct(int *qcoeffs, int *recoBlock, int qp, int cols, int rows,
	    int chroma, int mode)
{
  int
    size,			/* Size of block			*/
    coeffsTmpInt[64];

  size = cols*rows;

  newInvQuantCoefint(qcoeffs, coeffsTmpInt, size, qp, chroma, mode);
  varSizeInvDctFixedp(coeffsTmpInt, recoBlock, cols, rows);

}

/*
 *
 * decDctRegionInverseDct:
 *
 * Function:
 *      Inverse transforms and dequantizes a DCT block
 *
 * Returns:
 *      void
 */

void decInverseDct(int *diffBlk, int *dctCoeffs, int qp, int chroma,
		   int sizeX, int sizeY)
{

  int i, size, dctCoeffsTmp[64];

  size = sizeX*sizeY;

  if(size == 64)
    for(i = 0;i < size;i++)
      dctCoeffsTmp[i] = dctCoeffs[zig8x8[i]];
  else
    for(i = 0;i < size;i++)
      dctCoeffsTmp[i] = dctCoeffs[zig4x4[i]];

  decDct(dctCoeffsTmp, diffBlk, qp, sizeX, sizeY, chroma, INTER_BLOCK);

}
