/*
 * decDemux.h
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Contains:
 *    Functions for video demultiplexer.
 *
 */

#ifndef _DEC_BITS_H_
#define _DEC_BITS_H_

#include <stdio.h>

extern int maxDcLevels[32];
extern unsigned char
  highestDcLevBits[32];


#include "bib.h"
#include "nrctypes.h"

#define VDX_ERR_SYNC -2

#define VDX_OK_MERGE 1

#define VDX_OK_WITH_INTRA_PRED 0
#define VDX_OK_WITHOUT_INTRA_PRED 1

#define VDX_SNC_NO_SYNC 0
#define VDX_SNC_PSC 1
#define VDX_SNC_EOS 2

void init_get_bits();

#ifndef MAX_VLC_LEN
#define MAX_VLC_LEN 25
#endif
#define LEN_SIGNAL 255

typedef struct {
  bibBuffer_t *buffer;
} vdxInstance_t;

typedef struct
{
  u_char len;
  u_char code;
} vlc_item;

typedef struct 
{
  u_char index;
  u_char len;
} inv_vlc_item;

typedef struct 
{
  u_char last;
  u_char run;
  u_char level;
}ind2trip;

typedef struct 
{
  inv_vlc_item *table;
  u_char maxLength;
} vdxVlcToIndTbl_t;

typedef struct 
{
  u_char *table;
  u_char maxLength;  
} vdxVlcToPairIndTbl_t;

typedef struct 
{
  inv_vlc_item *table;
  u_char tableWidth;
} vdxVlcPairToIndTbl_t;

static inv_vlc_item invMCP_xy_P[1024];
static int two_powers[MAX_VLC_LEN];

extern bibBuffer_t *buff_in;

#define MODE_INTER 1
#define MODE_INTRA 2
#define MODE_COPY  3
#define MODE_INTRA_PRED         5
#define MODE_INTRA_PRED_VSCAN   6
#define MODE_INTRA_PRED_HSCAN   7
#define MODE_INTRA_FRAME        8
#define LAST    2
#define RUN    64
#define LEVEL 256


#ifndef BIBLIN
void vdxInitDecBits(const char *filename);
void vdxShutDecBits();
long int get_all_dec_bits();
#endif

/*
 * Global function prototypes.
 */

int vdxGetBits(int numberOfBits, bibBuffer_t *buffer, u_int32 *code_u32,
               vdxInstance_t *instance, int32 *bitError);

int32 vdxGetLongtermRVLC(int32 *index, vdxInstance_t *instance, 
                         int32 *bitError);

int vdxGetPicHead(int *IPB,          /* 0- P, 2- I    */ 
                  int *FrNo,
                  int *skip,         /* Distance from the previous frame    */
                  int *sliceMode,    /* Flag to signal that slice mode is on*/
                  int *useLoopFilter,/* Signals loopfilter usage            */
                  int *QP,
                  int *pic_sizex, int *pic_sizey,
                  int *numActiveRefFrames,/* Number of reference frames used*/
                  int32 *orderedRefFrames,/* Indexes of reference frames in
                                           * longterm memory                */
                  
                  /* memory location pointed by bits will be incremented
                   * by the number of bits read by this function
                   */
                  int32 *bits,
                  
                  /* 1- yes, one may not want to read PSC if sychronization
                   * was lost, and it has been recovered by detecting PCS
                   * by another routine
                   */
                  int ReadPSC,
                  vdxInstance_t *instance, int32 *bitError);

int vdxGetSliceHead(int *mbNum, int *intraSliceFlag, int *sliceQuant,
                    int *frNo, int *picSizeX, int *picSizeY, int readSSC,
                    int readISIandMBA, int *IPB,  /* 0- P, 2- I */
                    vdxInstance_t *instance, int32 *bitError);

int vdxOnlyZeroBits(vdxInstance_t *instance, int32 *bitError);

int vdxClose(vdxInstance_t *instance);

vdxInstance_t *vdxOpen(void);

int vdxCheckSync(int *numStuffBits, vdxInstance_t *instance);

int vdxCheckSSC(int *IPB,             /* 0- P, 2- I*/
                int *pic_sizex, int *pic_sizey,
                int *FrNo, int *skip, /* Distance from the previous frame    */
                int *QP, int32 *bits,
                vdxInstance_t *instance, int32 *bitError);

int vdxIntraGetMode(int32 *mode, vdxInstance_t *instance, int32 *bitError);

int vdxGetIntraFrameCBPY(int32 *index, vdxInstance_t *instance, 
                         int32 *bitError);

int vdxGetIntraFrameCBPC(int32 *index, vdxInstance_t *instance, 
                         int32 *bitError);

int vdxGet8bitsDC(int32 *level, vdxInstance_t *instance, int32 *bitError);

int vdxGetIntraPredDCTTriplet8x8(int32 *last, int32 *run, int32 *level, 
                                 int long_escape, vdxInstance_t *instance, 
                                 int32 *bitError);

int vdxGet8bitsDCPred(int *level, int *skipDc, int chroma, int size,
                      int prediction, vdxInstance_t *instance, 
                      int32 *bitError);

int vdxDctGetCncInfo(u_char *cnc_y, u_char *cnc_u, u_char *cnc_v,
                     int regionMode, int regionSize,vdxInstance_t *instance, 
                     int32 *bitError);

int vdxDctLoadDct(int *dctCoeffs, int size, int start,
                  vdxInstance_t *instance, int32 *bitError);

int vdxGetMotionPredInfo(int predDirsAvail, int predSelNeededUp,
                         int predSelNeededLeft, u_char *predDir, int *predSel,
                         vdxInstance_t *instance, int32 *bitError);

int vdxGetMotionCoeffs(int32 *A, int aQP, vdxInstance_t *instance,
                       int32 *bitError);


void vdxGetErrorCodingDcdTable(int32 blkSize, int32 mode, int32 chroma,
                               int32 acFlag, int32 qp, int32 *tableNo);

void indexToBitsErrorCodingDcd(int32 index, int32 chroma, int32 qp,
                               int32 *offsetDcdSubindex, 
                               int32 *bitsDcdSubindex);

int vdxGetRegionMode(int *regionMode, int *pecMode, int predPossible,
                     int regSize, vdxInstance_t *instance, int32 *bitError);

int32 vdxGetDctTable(int32 cols, int32 rows, int32 qp, int32 chroma, int32 mode,
                     int32 *tableNo, int32 *runBits, int32 *levelBits);

int32 vdxGetKltTable(int32 cols, int32 rows, int32 qp, int32 chroma,
                     int32 mode, int32 kltNum, int32 *tableNo, int32 *runBits,
                     int32 *levelBits);

void dcdRangeIndexToBits(int32 rangeIndex, int32 qp,
                         int32 *offsetDcdSubindex, int32 *bitsDcdSubindex);

void indexToBitsIntraBMV(int32 index, int32 *offsetBMVSubindex,
                         int32 *bitsBMVSubindex);



int32 vdxGetCodingMethod(vdxInstance_t *instance, int32 vlcTableInd, 
                         int32 *index, int32 *bitError);

int32 vdxGetTriplet(vdxInstance_t *instance, int32 vlcInd, 
                    int32 levelBits, int32 runBits, int32 *last, 
                    int32 *run, int32 *level, int32 *bitError);

int32 vdxGetIntraPredDcdCnc(vdxInstance_t *instance,
                            int32 qp, int32 predNo, int32 *predIndex,
                            int32 *dcdLevel, int32 *cnc, int32 *bitError);

int32 vdxGetIntraBMvector(int32 *bmVectY, int32 *bmVectX, int32 reducedSearchRange,
                          vdxInstance_t *instance, int32 *bitError);

int32 vdxGetCodingDcd(vdxInstance_t *instance, int32 vlcTableInd, 
                      int32 component, int32 qp, int32 *dcdLevel, 
                      int32 *bitError);

/*
 * DCM module prototypes
 */

int dcmRequestSliceRefill(bibBuffer_t **inBuffer, int *lastSlice);

#endif
