/*
 * fixedpDct.c
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Contains:
 *    Dct and idct functions
 *
 */


/*
 * There is a bug in Visual C 6,0 that sometimes prevents from compiling
 * source files that don't contain any compilable code. So we'll insert
 * something here.
 */
static int dummy;


#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "nrctypes.h"



/*
 * Number of fraction bits used for intermediate results.
 */
#define DCT4_PASS1_FRAC   5
#define DCT8_PASS1_FRAC   4


/*
 * If ROUND_INTERMEDIATE is defined, intermediate results are rounded.
 * If ROUND_INTERMEDIATE is not defined dct and idct are slightly faster.
 */

#define ROUND_INTERMEDIATE

/*
 * These macros convert fixed-point numbers to lower precision fized-point
 * numbers They perform correct rounding too.
 */
#ifdef ROUND_INTERMEDIATE
#define ROUND(a, bits) (short)( ((a) + (1<<((bits)-1))) >> (bits) )
#else
#define ROUND(a, bits) (short)( (a) >> (bits) )
#endif

#define ROUND2(a, bits) ( ((a) + (1<<((bits)-1))) >> (bits) )



/*
 * These define number of bits that are shifted after each of the passes.
 */
#define DCT4_PASS1_SHIFT (15 - DCT4_PASS1_FRAC)
#define DCT8_PASS1_SHIFT (15 - DCT8_PASS1_FRAC)

#define DCT4_PASS2_SHIFT (15 + DCT4_PASS1_FRAC)
#define DCT8_PASS2_SHIFT (15 + DCT8_PASS1_FRAC)


/*
 * Fixed-point DCT and IDCT constant
 */
#define DCT4_COEF_A 16384        /* sqrt(1.0/4.0)                     */
#define DCT4_COEF_B 21407        /* sqrt(1.0/2.0) * cos(1.0*M_PI/8.0) */
#define DCT4_COEF_D 8867         /* sqrt(1.0/2.0) * cos(3.0*M_PI/8.0) */

#define DCT8_COEF_A 11585        /* sqrt(1.0/8.0)                      */
#define DCT8_COEF_B 16069        /* sqrt(1.0/4.0) * cos(1.0*M_PI/16.0) */
#define DCT8_COEF_C 15137        /* sqrt(1.0/4.0) * cos(2.0*M_PI/16.0) */
#define DCT8_COEF_D 13623        /* sqrt(1.0/4.0) * cos(3.0*M_PI/16.0) */
#define DCT8_COEF_F 9102         /* sqrt(1.0/4.0) * cos(5.0*M_PI/16.0) */
#define DCT8_COEF_G 6270         /* sqrt(1.0/4.0) * cos(6.0*M_PI/16.0) */
#define DCT8_COEF_H 3196         /* sqrt(1.0/4.0) * cos(7.0*M_PI/16.0) */




/******************************************************************************
 *
 *                DCT and IDCT macros and functions
 *
 *****************************************************************************/


#define DCT8_MACRO_NORMAL(SI, SJ, DEST, RNDMAC, SHIFT)          \
{                                                               \
  int                                                           \
    tmp10, tmp11, tmp12, tmp13;                                 \
  tmp10 = tmp0 + tmp6; \
  tmp11 = tmp2 + tmp4; \
  tmp12 = tmp0 - tmp6; \
  tmp13 = tmp2 - tmp4; \
  DEST[SI*i+SJ*0] = RNDMAC(DCT8_COEF_A*tmp10 + DCT8_COEF_A*tmp11, (SHIFT)); \
  DEST[SI*i+SJ*2] = RNDMAC(DCT8_COEF_C*tmp12 + DCT8_COEF_G*tmp13, (SHIFT)); \
  DEST[SI*i+SJ*4] = RNDMAC(DCT8_COEF_A*tmp10 - DCT8_COEF_A*tmp11, (SHIFT)); \
  DEST[SI*i+SJ*6] = RNDMAC(DCT8_COEF_G*tmp12 - DCT8_COEF_C*tmp13, (SHIFT)); \
}

#define DCT8_MACRO_SLOW(SI, SJ, DEST, RNDMAC, SHIFT) \
  DEST[SI*i+SJ*0] = RNDMAC(DCT8_COEF_A*tmp0 + DCT8_COEF_A*tmp2 + DCT8_COEF_A*tmp4 + DCT8_COEF_A*tmp6, (SHIFT)); \
  DEST[SI*i+SJ*2] = RNDMAC(DCT8_COEF_C*tmp0 + DCT8_COEF_G*tmp2 - DCT8_COEF_G*tmp4 - DCT8_COEF_C*tmp6, (SHIFT)); \
  DEST[SI*i+SJ*4] = RNDMAC(DCT8_COEF_A*tmp0 - DCT8_COEF_A*tmp2 - DCT8_COEF_A*tmp4 + DCT8_COEF_A*tmp6, (SHIFT)); \
  DEST[SI*i+SJ*6] = RNDMAC(DCT8_COEF_G*tmp0 - DCT8_COEF_C*tmp2 + DCT8_COEF_C*tmp4 - DCT8_COEF_G*tmp6, (SHIFT));

/* TYPE is used to choose between two macros: DCT8_MACRO_NORMAL and DCT8_MACRO_SLOW */
#define DCT8_MACRO(TYPE, LINES, SI, SJ, SRC, DEST, RNDMAC, SHIFT) \
{                                                               \
  int                                                           \
    i, tmp0, tmp1, tmp2, tmp3, tmp4, tmp5, tmp6, tmp7;          \
  for (i = 0; i < LINES; i++) {                                 \
    tmp0 = SRC[SI*i+SJ*0] + SRC[SI*i+SJ*7]; \
    tmp2 = SRC[SI*i+SJ*1] + SRC[SI*i+SJ*6]; \
    tmp4 = SRC[SI*i+SJ*2] + SRC[SI*i+SJ*5]; \
    tmp6 = SRC[SI*i+SJ*3] + SRC[SI*i+SJ*4]; \
    tmp1 = SRC[SI*i+SJ*0] - SRC[SI*i+SJ*7]; \
    tmp3 = SRC[SI*i+SJ*1] - SRC[SI*i+SJ*6]; \
    tmp5 = SRC[SI*i+SJ*2] - SRC[SI*i+SJ*5]; \
    tmp7 = SRC[SI*i+SJ*3] - SRC[SI*i+SJ*4]; \
    DCT8_MACRO_ ## TYPE (SI, SJ, DEST, RNDMAC, SHIFT) \
    DEST[SI*i+SJ*1] = RNDMAC(DCT8_COEF_B*tmp1 + DCT8_COEF_D*tmp3 + DCT8_COEF_F*tmp5 + DCT8_COEF_H*tmp7, (SHIFT)); \
    DEST[SI*i+SJ*3] = RNDMAC(DCT8_COEF_D*tmp1 - DCT8_COEF_H*tmp3 - DCT8_COEF_B*tmp5 - DCT8_COEF_F*tmp7, (SHIFT)); \
    DEST[SI*i+SJ*5] = RNDMAC(DCT8_COEF_F*tmp1 - DCT8_COEF_B*tmp3 + DCT8_COEF_H*tmp5 + DCT8_COEF_D*tmp7, (SHIFT)); \
    DEST[SI*i+SJ*7] = RNDMAC(DCT8_COEF_H*tmp1 - DCT8_COEF_F*tmp3 + DCT8_COEF_D*tmp5 - DCT8_COEF_B*tmp7, (SHIFT)); \
  } \
}

#define IDCT8_MACRO(LINES, SI, SJ, SRC, DEST, RNDMAC, SHIFT)    \
{                                                               \
  int                                                           \
    i, tmp0, tmp1, tmp2, tmp3, tmp4, tmp5, tmp6, tmp7,          \
    tmp10, tmp11, tmp12, tmp13;                                 \
  for (i = 0; i < LINES; i++) {                                 \
    tmp0 = DCT8_COEF_A*SRC[SI*i+SJ*0] + DCT8_COEF_A*SRC[SI*i+SJ*4]; \
    tmp1 = DCT8_COEF_A*SRC[SI*i+SJ*0] - DCT8_COEF_A*SRC[SI*i+SJ*4]; \
    tmp2 = DCT8_COEF_C*SRC[SI*i+SJ*2] + DCT8_COEF_G*SRC[SI*i+SJ*6]; \
    tmp3 = DCT8_COEF_G*SRC[SI*i+SJ*2] - DCT8_COEF_C*SRC[SI*i+SJ*6]; \
    tmp4 = DCT8_COEF_B*SRC[SI*i+SJ*1] + DCT8_COEF_D*SRC[SI*i+SJ*3] + DCT8_COEF_F*SRC[SI*i+SJ*5] + DCT8_COEF_H*SRC[SI*i+SJ*7]; \
    tmp5 = DCT8_COEF_D*SRC[SI*i+SJ*1] - DCT8_COEF_H*SRC[SI*i+SJ*3] - DCT8_COEF_B*SRC[SI*i+SJ*5] - DCT8_COEF_F*SRC[SI*i+SJ*7]; \
    tmp6 = DCT8_COEF_F*SRC[SI*i+SJ*1] - DCT8_COEF_B*SRC[SI*i+SJ*3] + DCT8_COEF_H*SRC[SI*i+SJ*5] + DCT8_COEF_D*SRC[SI*i+SJ*7]; \
    tmp7 = DCT8_COEF_H*SRC[SI*i+SJ*1] - DCT8_COEF_F*SRC[SI*i+SJ*3] + DCT8_COEF_D*SRC[SI*i+SJ*5] - DCT8_COEF_B*SRC[SI*i+SJ*7]; \
    tmp10 = tmp0 + tmp2; \
    tmp11 = tmp1 + tmp3; \
    tmp12 = tmp1 - tmp3; \
    tmp13 = tmp0 - tmp2; \
    DEST[SI*i+SJ*0] = RNDMAC(tmp10 + tmp4, (SHIFT)); \
    DEST[SI*i+SJ*1] = RNDMAC(tmp11 + tmp5, (SHIFT)); \
    DEST[SI*i+SJ*2] = RNDMAC(tmp12 + tmp6, (SHIFT)); \
    DEST[SI*i+SJ*3] = RNDMAC(tmp13 + tmp7, (SHIFT)); \
    DEST[SI*i+SJ*4] = RNDMAC(tmp13 - tmp7, (SHIFT)); \
    DEST[SI*i+SJ*5] = RNDMAC(tmp12 - tmp6, (SHIFT)); \
    DEST[SI*i+SJ*6] = RNDMAC(tmp11 - tmp5, (SHIFT)); \
    DEST[SI*i+SJ*7] = RNDMAC(tmp10 - tmp4, (SHIFT)); \
  } \
}

#define DCT4_MACRO(LINES, SI, SJ, SRC, DEST, RNDMAC, SHIFT)     \
{                                                               \
  int                                                           \
    i, tmp0, tmp1, tmp2, tmp3;                                  \
  for (i = 0; i < LINES; i++) {                                 \
    tmp0 = SRC[SI*i+SJ*0] + SRC[SI*i+SJ*3]; \
    tmp1 = SRC[SI*i+SJ*1] + SRC[SI*i+SJ*2]; \
    tmp2 = SRC[SI*i+SJ*0] - SRC[SI*i+SJ*3]; \
    tmp3 = SRC[SI*i+SJ*1] - SRC[SI*i+SJ*2]; \
    DEST[SI*i+SJ*0] = RNDMAC(DCT4_COEF_A*tmp0 + DCT4_COEF_A*tmp1, (SHIFT)); \
    DEST[SI*i+SJ*1] = RNDMAC(DCT4_COEF_B*tmp2 + DCT4_COEF_D*tmp3, (SHIFT)); \
    DEST[SI*i+SJ*2] = RNDMAC(DCT4_COEF_A*tmp0 - DCT4_COEF_A*tmp1, (SHIFT)); \
    DEST[SI*i+SJ*3] = RNDMAC(DCT4_COEF_D*tmp2 - DCT4_COEF_B*tmp3, (SHIFT)); \
  } \
}

#define IDCT4_MACRO(LINES, SI, SJ, SRC, DEST, RNDMAC, SHIFT)    \
{                                                               \
  int                                                           \
    i, tmp0, tmp1, tmp2, tmp3;                                  \
  for (i = 0; i < LINES; i++) {                                 \
    tmp0 = DCT4_COEF_A*SRC[SI*i+SJ*0] + DCT4_COEF_A*SRC[SI*i+SJ*2]; \
    tmp1 = DCT4_COEF_A*SRC[SI*i+SJ*0] - DCT4_COEF_A*SRC[SI*i+SJ*2]; \
    tmp2 = DCT4_COEF_B*SRC[SI*i+SJ*1] + DCT4_COEF_D*SRC[SI*i+SJ*3]; \
    tmp3 = DCT4_COEF_D*SRC[SI*i+SJ*1] - DCT4_COEF_B*SRC[SI*i+SJ*3]; \
    DEST[SI*i+SJ*0] = RNDMAC(tmp0 + tmp2, (SHIFT)); \
    DEST[SI*i+SJ*1] = RNDMAC(tmp1 + tmp3, (SHIFT)); \
    DEST[SI*i+SJ*2] = RNDMAC(tmp1 - tmp3, (SHIFT)); \
    DEST[SI*i+SJ*3] = RNDMAC(tmp0 - tmp2, (SHIFT)); \
  } \
}


void dct8x8( int *source, int *destination )
{
  short
    tmp[8*8];

  /* Horizintal & vertical 1D-dct */

  DCT8_MACRO(NORMAL, 8, 8, 1, source, tmp, ROUND, DCT8_PASS1_SHIFT);
  DCT8_MACRO(SLOW, 8, 1, 8, tmp, destination, ROUND2, DCT8_PASS2_SHIFT);
}


void invDct8x8( int *source, int *destination )
{
  short
    tmp[8*8];

  /* Horizintal & vertical 1D-dct */
  IDCT8_MACRO(8, 8, 1, source, tmp, ROUND, DCT8_PASS1_SHIFT);
  IDCT8_MACRO(8, 1, 8, tmp, destination, ROUND2, DCT8_PASS2_SHIFT);
}


void dct8x4( int *source, int *destination )
{
  short
    tmp[8*4];

  /* Horizintal & vertical 1D-dct */
  DCT8_MACRO(NORMAL, 4, 8, 1, source, tmp, ROUND, DCT8_PASS1_SHIFT);
  DCT4_MACRO(8, 1, 8, tmp, destination, ROUND2, DCT8_PASS2_SHIFT);
}

void invDct8x4( int *source, int *destination )
{
  short
    tmp[8*4];

  /* Horizintal & vertical 1D-idct */
  IDCT8_MACRO(4, 8, 1, source, tmp, ROUND, DCT8_PASS1_SHIFT);
  IDCT4_MACRO(8, 1, 8, tmp, destination, ROUND2, DCT8_PASS2_SHIFT);
}

void dct4x8( int *source, int *destination )
{
  short
    tmp[4*8];

  /* Horizintal & vertical 1D-dct */
  DCT4_MACRO(8, 4, 1, source, tmp, ROUND, DCT8_PASS1_SHIFT);
  DCT8_MACRO(NORMAL, 4, 1, 4, tmp, destination, ROUND2, DCT8_PASS2_SHIFT);
}

void invDct4x8( int *source, int *destination )
{
  short
    tmp[4*8];

  /* Horizintal & vertical 1D-idct */
  IDCT4_MACRO(8, 4, 1, source, tmp, ROUND, DCT8_PASS1_SHIFT);
  IDCT8_MACRO(4, 1, 4, tmp, destination, ROUND2, DCT8_PASS2_SHIFT);
}

void dct4x4( int *source, int *destination )
{
  short
    tmp[4*4];

  /* Horizintal & vertical 1D-dct */
  DCT4_MACRO(4, 4, 1, source, tmp, ROUND, DCT4_PASS1_SHIFT);
  DCT4_MACRO(4, 1, 4, tmp, destination, ROUND2, DCT4_PASS2_SHIFT);
}

void invDct4x4( int *source, int *destination )
{
  short
    tmp[4*4];

  /* Horizintal & vertical 1D-idct */
  IDCT4_MACRO(4, 4, 1, source, tmp, ROUND, DCT4_PASS1_SHIFT);
  IDCT4_MACRO(4, 1, 4, tmp, destination, ROUND2, DCT4_PASS2_SHIFT);
}



/*****************************************************************************
 *
 * Function:     void varSizeDctFixedp(int *source,int *destination,
 *                                     int cols,int rows)
 *
 * Description:  Variably-sized ([4,8]x[4,8]) 2D DCT, non-MMX version
 *
 * Input:        short *source           Source coefficients
 *               short *destination      Destination buffer
 *               int cols                Number of columns (must be 4 or 8)
 *               int rows                Number of rows (must be 4 or 8)
 *
 ****************************************************************************/
void varSizeDctFixedp( int *source, int *destination, int cols, int rows )
{
  assert((cols == 4 || cols == 8) && (rows == 4 || rows == 8));

  if (cols == 4) {
    if (rows == 4)
      dct4x4(source, destination);
    else
      dct4x8(source, destination);
  }
  else {
    if (rows == 4)
      dct8x4(source, destination);
    else
      dct8x8(source, destination);
  }
}


/*****************************************************************************
 *
 * Function:     void varSizeInvDctFixedp(int *source,int *destination,
 *                                        int cols,int rows)
 *
 * Description:  Variably-sized ([4,8]x[4,8]) 2D inverse DCT, non-MMX version
 *
 * Input:        short *source           Source coefficients
 *               short *destination      Destination buffer
 *               int cols                Number of columns (must be 4 or 8)
 *               int rows                Number of rows (must be 4 or 8)
 *
 ****************************************************************************/
void varSizeInvDctFixedp( int *source, int *destination, int cols, int rows )
{
  assert((cols == 4 || cols == 8) && (rows == 4 || rows == 8));

  if (cols == 4) {
    if (rows == 4)
      invDct4x4(source, destination);
    else
      invDct4x8(source, destination);
  }
  else {
    if (rows == 4)
      invDct8x4(source, destination);
    else
      invDct8x8(source, destination);
  }
}


