#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>

#include "mainFlags.h"
#include "tools.h"
#include "decDemux.h"
#include "encMux.h"
#include "deblockFilt.h"
#include "IOimg.h"
#include "IntraPredDecoder.h"
#include "debug.h"
#include "PredErrorCoder.h"
#include "PredErrorDecoder.h"
#include "toolMCC.h"
#include "encMCC.h"
#include "decMCC.h"

/*
 * Global counter used in dumping of TCOEFF information.
 */
extern int
  triplet_counter_Y,
  intra_dc_counter_Y,
  doublet_counter_Y,
  vq_counter_Y,
  *region_intra_dc,
  triplet_counter,
  intra_dc_counter,
  doublet_counter,
  vq_counter,
  rdc_counter,
  *rdc_value;
extern doublet
  *region_doublets;
extern triplet
  *region_triplets,
  *vq_triplets;

/*
 * This table defines the chroma QP in a similar
 * fashion used in H.263+ Annex T.
 */
extern u_char
  annexTlookup[32];

/*
 * Local function used for BM search in intra prediction.
 * Currently at run time, the BM is not among the intra
 * prediciton set of methods used but this may change soon.
 */
static int32 findIntraMotion(int32 *origBlk, u_char *recoImag, int32 *predBlk,
                             int32 x, int32 y, int32 *motNo, int32 blocksizex,
                             int32 blocksizey, int32 x_size, int32 prevMin,
                             int32 lambda, int32 *bits, int reducedSearchRange,
                             int32 noOfValidVectors, int32 *validX, int32 *validY);


/***************************************************************************
 * Function    : codeIntraBlock()
 *
 * Purpose     : Encodes and dumps the block layer bitstream for an Intra
 *               block (either luma or chroma). Intra prediction is used
 *               followed by MCC prediction error coding.
 *
 * Arguments   : See the function call
 *
 * Returns     : void
 *
 *               Copyright (c) 1997 Nokia Research Center
 ***************************************************************************/

static void codeIntraBlock(Byte *origFr,         /* In:  Current (original) frame  */
                           Byte *recoFr,         /* Out: Reconstruction frame buf  */
                           int16 *dcFrameBuf,    /* Out: Frame buf for DC pred     */
                           u_char *classifBuf,   /* Out: Frame buf for dir classif */
                           int *acBits,          /* Out: Bits for intra pred error */
                           int *ovhBits,         /* Out: Bits for intra pred info  */
                           int mode,             /* In:  Flag to signal I / P frame*/
                           int chroma,           /* In:  Flag to signal luma/chroma*/
                           int qp,               /* In:  Quant parameter to use    */
                           int lambda,           /* In:  Lambda parameter to use   */
                           int maxDClevels,      /* In:  Max # of DC quant levels  */
                           int blkSizex,         /* In:  Block height              */
                           int blkSizey,         /* In:  Block width               */
                           int frX,              /* In:  Block x coord inside frame*/
                           int frY,              /* In:  Block y coord inside frame*/
                           int frStepx,          /* In:  8 for luma, 4 for chroma  */
                           int frStepy,          /* In:  8 for luma, 4 for chroma  */
                           int frSizex,          /* In:  Frame width # of pels     */
                           int frSizey,          /* In:  Frame height # of pels    */
                           int xBlocks,          /* In:  Frame width # of blocks   */
                           int gen_bitstr_flag,  /* In:  Dump output to bitstream  */
                           int sliceXmin,        /* In:  Xmin  of current slice    */
                           int sliceXmax,        /* In:  Xmax  of current slice    */
                           int sliceYmin,        /* In:  Ymin  of current slice    */
                           int sliceYmax,        /* In:  Ymax  of current slice    */
                           int16 *slicePredMap,  /* In:  Slice label buffer        */
                           int sliceSx,          /* In:  Horizontal slice buf size */
                           int sliceSy,          /* In:  Vertical slice buf size   */
                           int currSliceID,      /* In:  Label of current slice    */
                           int complexityMode)   /* In:  Level of encod complexity */
{
  /*
   * Various constants used inside this functions
   */
  #define IP_METH_DC                  1
  #define IP_METH_BM                 13
  #define MIN_PRED_ERR_BITS           4
  #define ACCEPTABLE_MSE            7/4
  #define HUGE_LAGRANGIAN_COST  5000000

  int
    maxValue = 255, maxLevel = 256, rotateFlag,
    methods = AMOUNT_OF_8x8_METH+1,
    leftEdge, upperEdge, leftDc, upperDc, upleftDc,
    blkSize, frBlkNo, predNo, tmpLeft, tmpUpper,
    dcOrigSum, dcOrig, indOrig, methName, classTmp,
    prevMin, bestLagr, bestMeth, bestCode,
    bestDiff, bestCom, bestMot, bestCnc, bestNo,
    comBit, reducedSearchRange,
    class8x8, dcPredSum, dcdIndex,
    dcDiff, ncCom, codCom,
    cncDif, seBefore, lagr, prevLagr,
    codMeth8x8, blkClass, codMeth,
    cbsBits, dctBits, mesBits, seAfter, pels, class4x4,
    codMeth4x4, coded, acBit, bits, bestCM,
    bestAc, bestErro, decBits, dcRecoSum,
    addy, addx, addNo, tmpInd, methRank,
    origBlk[64], predBlk[64], diffBlk[64], recoBlk[64],
    methCoeffs[64], dctCoeffs[64], ExtBlk[16], tmpBlk[64],
    tmpPredBlk[64], tmpDiffBlk[64], dct4x4Meth, bestBmv,
    i, k, bestValid,
    methPtr[MAX_IP_LIST_SIZE],
    maxNoOfValidVectors, noOfValidVectors, totalNoOfVectors,
    bestPredDir, bestDiffDir, bestEnergy,
    tmpThres, tmpPredNo, numOfIntraPredMeths,
    selNPtr[MAX_IP_LIST_SIZE];

  u_char
    origNo = UCHAR_MAX, predDirIn, diffDirIn; 

  int32
    motNo, motBit,
    validI[AMOUNT_OF_INTRA_MOTION],
    validX[AMOUNT_OF_INTRA_MOTION],
    validY[AMOUNT_OF_INTRA_MOTION];

  blkSize = blkSizex*blkSizey;
  frBlkNo = (frY/frStepy)*xBlocks + (frX/frStepx);

  /* 
   * Depending on whether the block at the
   * left and the block at the right are inside
   * the same slice as the current block the
   * variables leftEdge and upperEdge are set
   * to contain the classifier directionality
   * of the corresponding left and right block.
   *
   * The variables leftDc, upperDc and upleftDc
   * are set to be holding information about the
   * DC level of the neighboring blocks.
   *
   * In case the neighboring blocks are
   * outside the curent slice an invalid
   * value of -1 or UCHAR_MAX is assigned
   * for the sake of simplicity.
   */

  if(frX >= frStepx && frY >= frStepy) {
    if (slicePredMap[frBlkNo-1] == currSliceID) {
      leftEdge  = classifBuf[frBlkNo-1];
      leftDc    = dcFrameBuf[frBlkNo-1];
    }
    else {
      leftEdge  = UCHAR_MAX;
      leftDc    = -1;
    }
    if (slicePredMap[frBlkNo-xBlocks] == currSliceID) {
      upperEdge = classifBuf[frBlkNo-xBlocks];
      upperDc   = dcFrameBuf[frBlkNo-xBlocks];
    }
    else {
      upperEdge = UCHAR_MAX;
      upperDc   = -1;
    }
    if (slicePredMap[frBlkNo-xBlocks-1] == currSliceID)
      upleftDc  = dcFrameBuf[frBlkNo-xBlocks-1];
    else
      upleftDc  = -1;
  }
  else if(frX >= frStepx) {
    if (slicePredMap[frBlkNo-1] == currSliceID) {
      leftEdge  = classifBuf[frBlkNo-1];
      leftDc    = dcFrameBuf[frBlkNo-1];
    }
    else {
      leftEdge  = UCHAR_MAX;
      leftDc    = -1;
    }
    upperEdge = UCHAR_MAX;
    upperDc   = -1;
    upleftDc  = -1;
  }
  else if(frY >= frStepy) {
    if (slicePredMap[frBlkNo-xBlocks] == currSliceID) {
      upperEdge = classifBuf[frBlkNo-xBlocks];
      upperDc   = dcFrameBuf[frBlkNo-xBlocks];
    }
    else {
      upperEdge = UCHAR_MAX;
      upperDc   = -1;
    }
    leftEdge  = UCHAR_MAX;
    leftDc    = -1;
    upleftDc  = -1;
  }
  else {
    leftEdge  = UCHAR_MAX;
    upperEdge = UCHAR_MAX;
    leftDc    = -1;
    upperDc   = -1;
    upleftDc  = -1;
  }

  /*
   * Get the original blkSizex x blkSizey block
   * from frame buffer "origFr" into block buffer "origBlk".
   * The possible block sizes are 8x8 or 4x4.
   */
  getIntBlkFromByteImage(origFr, origBlk, blkSizex, blkSizey,
                         frX, frY, frSizex);

  /*
   * Get the quantized DC value of the current original block.
   */

  dcOrigSum = 0;
  for(i=0; i<blkSize; i++)
    dcOrigSum += origBlk[i];

  dcOrig = ipdDCquantAndInvQuant(dcOrigSum, blkSize, maxValue, maxDClevels, &indOrig);

  /*
   * Mapping of directional classes (produced
   * by the gradient based cllassifier) into 7
   * directions corresponding to:
   *  0 - horizontal
   *  1 - diagonal /
   *  2 - vertical
   *  3 - diagonal \
   *  4 - flat
   *  5 - texture
   *  6 - block outside current slice or frame
   */
  if(leftEdge == UCHAR_MAX)  tmpLeft = 6;
  else if(leftEdge == N_0_8) tmpLeft = 0;
  else if(leftEdge == N_1_8) tmpLeft = 1;
  else if(leftEdge == N_2_8) tmpLeft = 1;
  else if(leftEdge == N_3_8) tmpLeft = 1;
  else if(leftEdge == N_4_8) tmpLeft = 2;
  else if(leftEdge == N_5_8) tmpLeft = 3;
  else if(leftEdge == N_6_8) tmpLeft = 3;
  else if(leftEdge == N_7_8) tmpLeft = 3;
  else if(leftEdge == FLAT0) tmpLeft = 4;
  else if(leftEdge == TEXT1) tmpLeft = 5;
  else if(leftEdge == TEXT2) tmpLeft = 5;
  else {
    printf("Wrong value leftEdge %d in codeIntraBlock()\n", leftEdge);
    exit(1);
  }
  if(upperEdge == UCHAR_MAX)  tmpUpper = 6;
  else if(upperEdge == N_0_8) tmpUpper = 0;
  else if(upperEdge == N_1_8) tmpUpper = 1;
  else if(upperEdge == N_2_8) tmpUpper = 1;
  else if(upperEdge == N_3_8) tmpUpper = 1;
  else if(upperEdge == N_4_8) tmpUpper = 2;
  else if(upperEdge == N_5_8) tmpUpper = 3;
  else if(upperEdge == N_6_8) tmpUpper = 3;
  else if(upperEdge == N_7_8) tmpUpper = 3;
  else if(upperEdge == FLAT0) tmpUpper = 4;
  else if(upperEdge == TEXT1) tmpUpper = 5;
  else if(upperEdge == TEXT2) tmpUpper = 5;
  else {
    printf("Wrong value upperEdge %d in codeIntraBlock()\n", upperEdge);
    exit(1);
  }

  /*
   * The context provided by Left and Upper blocks
   * is signaled by predNo. This variable
   * can take 7 x 7 = 49 posible values.
   */
  predNo = tmpLeft*7 + tmpUpper;

  /*
   * The table ipContextLists[] contains
   * the context dependent lists of intra
   * prediction methods. Its size is
   * 49 x (MAX_IP_LIST_SIZE+1). Each of the
   * 49 rows corresponds to a (L, U) context.
   * In ipContextLists[] the first column
   * represents the length of each list which
   * is here copied into variable numOfIntraPredMeths.
   */
  tmpPredNo = predNo*(MAX_IP_LIST_SIZE+1);
  numOfIntraPredMeths = ipContextLists[tmpPredNo];

  /*
   * Since the first column in ipContextLists[]
   * represents the length of each list this auxiliary
   * variable tmpPredNo is incremented to point to the
   * head of intra pred methods list.
   */
  tmpPredNo++;

  /* Extract the corresponding list from ipContextLists[]
   * into the buffer methPtr[]. The ranks "k" of each method
   * inside the list is also copied into a buffer selNPtr[].
   */
  for(k=0; k<numOfIntraPredMeths; k++) {
    methPtr[k] = ipContextLists[tmpPredNo + k];
    selNPtr[k] = k;
  }

  /*
   * Initialize loop variables used when inspecting the
   * intra prediction methods in methPtr[] one by one.
   */

  prevMin     = HUGE_LAGRANGIAN_COST;
  bestLagr    = HUGE_LAGRANGIAN_COST;
  bestMeth    = -1;
  bestCode    = -1;
  bestDiff    = -500;
  bestCom     = -1;
  bestMot     = -1;
  bestCnc     = -1;
  bestNo      = -1;
  bestBmv     = 0;
  bestValid   = 0;
  bestPredDir = -1;
  bestDiffDir = -1;
  bestEnergy  = -1;

  for(k=0; k<numOfIntraPredMeths; k++)
  {
    /*
     * Get the current intra prediction method in methName.
     * Get its rank (inside the context dependent list) in methRank.
     */
    methName  = methPtr[k];
    methRank  = selNPtr[k];

    motNo     =  0;
    noOfValidVectors  =  0;
    predDirIn = UCHAR_MAX;
    diffDirIn = UCHAR_MAX;

    /*
     * The BM intra prediction method is not allowed
     * for 4x4 block sizes.
     */
    if(methName == IP_METH_BM && blkSize == SIZE_4x4)
      continue;
    else
    {
      if (methName == IP_METH_BM)
        motBit = SendBMV_Event(0, -8, 1, 0);
      else
        motBit = 0;

      lagr = lambda *
             ((comBit = SendSelDcdCncSign_Event(predNo, methRank, 0, 0, 0, qp, 0)) + motBit);

      class8x8 = methName + (NON_CLASSIFIED+1);

      if(lagr < bestLagr)
      {
        /* Perform Block Matching for intra prediction */

        if(methName == IP_METH_BM)
        {
          /*
           * For chroma in I frames or for any intra blocks
           * in P frames the BM search quits when 136/320 valid
           * vectors are found. A valid vector must point to
           * a block which is fully inside current slice.
           *
           * For luma in I frames up to 320 valid vectors
           * are allowed.
           */
          if(chroma == CHROMA || mode == INTRA_BLOCK)
          {
            totalNoOfVectors = maxNoOfValidVectors = 136;
            reducedSearchRange  = 1;
          }
          else
          {
            totalNoOfVectors = maxNoOfValidVectors = AMOUNT_OF_INTRA_MOTION;
            reducedSearchRange  = 0;
          }

          /*
           * The search quits when above maxNoOfValidVectors is reached
           * OR when all the 512 vectors have been scanned. The following
           * function finds the set of valid vectors and places their
           * offsets in validX[], validY[].
           */
          noOfValidVectors = findValidVectors(validX, validY, validI, maxNoOfValidVectors,
                                              totalNoOfVectors, frX, frY, blkSizex, blkSizey,
                                              frSizex, sliceXmin, sliceXmax, sliceYmin,
                                              sliceYmax, slicePredMap, sliceSx,
                                              sliceSy, currSliceID, mode);

          /*
           * Choose the best valid vector.
           */
          dcPredSum = findIntraMotion((int32*)origBlk, recoFr,
                                      (int32*)predBlk, frX, frY, &motNo,
                                      blkSizex, blkSizey, frSizex, prevMin,
                                      lambda, &motBit, reducedSearchRange,
                                      noOfValidVectors, validX, validY);
        }
        else if(methName < 1 || methName > 12) {
          printf("Wrong methName %d in codeIntraBlock()\n", methName);
          exit(1);
        }
        else
        {
          /* Any intra prediction method which is not Block Matching */

          /*
           * Perform DC prediction (if methName is IP_METH_DC) or
           * one of the 11 directional prediction (if methName is
           * otherwise). The IP_METH_BM was dealt with just above.
           * If the given prediction method cannot be performed
           * such when at the border of the slice the return value
           * is negative. Otherwise, it's the sum of pred values.
           */
          dcPredSum = predIntra(recoFr, (int32*)predBlk, frX, frY, methName,
                                blkSizex, blkSizey, frSizex, frSizey,
                                sliceXmin, sliceXmax, sliceYmin, sliceYmax,
                                leftDc, upperDc, upleftDc, slicePredMap,
                                sliceSx, sliceSy, currSliceID);
        }

        /*
         * In case the intra prediction cannot be performed
         * (such as when doing BM at border of slice) then
         * dcPredSum is returned negative.
         */

        if(dcPredSum < 0)
          continue;
        else
        {
          /*
           * The difference sum is quantised and normalised to
           * the block size. In the followings, an iteration
           * is performed only in the case when clipping of
           * pixel values occurs.
           */

          int
            diffTmp   = 0,
            dcDiffSum = 0,
            diffFlag  = 1,
            diffSum   = dcOrigSum - dcPredSum;

          while(diffFlag)
          {
            if(blkSize-diffTmp > 0)
              dcDiffSum += (diffSum*blkSize)/(blkSize-diffTmp);
            else {
              fprintf(stderr, "Error: blkSize %d or diffTmp %d wrong\n",
                      blkSize, diffTmp);
              exit(1);
            }

            /*
             * The difference sum is quantised and normalised to
             * the block size. The inverse quantisation value is
             * returned into the variable dcDiff.
             */

            dcDiff = ipdDCquantAndInvQuant(dcDiffSum, blkSize, maxValue,
                                           maxDClevels, &dcdIndex);

            diffTmp = 0;
            diffSum = 0;

            /*
             * Intra prediction block values + DC are saturated to
             * 0, 255 range and placed in predBlk[]. The intra prediction
             * error block is prepared in diffBlk[] block buffer.
             */

            for(i=0; i<blkSize; i++)
            {
              tmpBlk[i] = predBlk[i] + dcDiff;
              if(tmpBlk[i] > 255) {
                tmpBlk[i] = 255;
                diffTmp++;
              }
              else if(tmpBlk[i] < 0) {
                tmpBlk[i] = 0;
                diffTmp++;
              }
              diffSum += (origBlk[i] - tmpBlk[i]);
            }

            if(diffTmp == blkSize || diffTmp == 0)
              diffFlag = 0;
            else
            {
              if((dcDiff < 0 && diffSum > 0) ||
                 (dcDiff > 0 && diffSum < 0))
                diffFlag = 0;
              else
                ipdDCquantAndInvQuant(diffSum, blkSize-diffTmp, maxValue,
                                      maxDClevels, &diffFlag);
            }
          }

          for(i=0; i<blkSize; i++)
          {
            predBlk[i] = tmpBlk[i];
            diffBlk[i] = origBlk[i] - predBlk[i];
          }

          /*
           * Calculate the minimum cost (in bits) of signaling
           * intra prediction method, plus BM vector (if any),
           * plus DC difference index, and coded not coded bit.
           * If the sub-term (lambda * bits) already
           * exceeds bestLagr so far then one can just forget
           * about current intra prediction method and proceed
           * to the next one in the list.
           */

          lagr = lambda *
                 ((ncCom = SendSelDcdCncSign_Event(predNo, methRank,
                                                   abs(dcdIndex), 0,
                                                   ((dcdIndex < 0) ? -1 : 1),
                                                   qp, 0)) + motBit);
          comBit = codCom = ncCom;
          cncDif = 0;

          /*
           * If best lagrangian so far is larger than the
           * default component of current lagrangian then
           * try encoding the current intra prediction error.
           */
          if (lagr < bestLagr)
          {
            /*
             * This is the error energy before
             * prediction error is applied.
             */
            seBefore = calcBlockEnergy(diffBlk, blkSize);

            /*
             * A test for fast estimation whether coding of
             * prediction error helps against not coded option.
             */
            if(seBefore > MIN_PRED_ERR_BITS * lambda + blkSize * ACCEPTABLE_MSE)
            {
              /*
               * Deduce the component in lagrangian which corresponds to
               * intra prediction because MCC uses some previous best
               * lagrangian in its search for best coding method.
               */
              prevLagr = bestLagr - lagr;

              /*
               * Intra prediction error coding starts.
               */
              if(blkSize == SIZE_8x8)
              {
                /*
                 * If BM was uses as intra prediction then
                 * the predicted texture is directionality classified
                 * in order to select the set of coding methods.
                 */
                if(methName == IP_METH_BM)
                  blockEdge(predBlk, blkSizex, blkSizey, &predDirIn);
                else
                  predDirIn = TEXT2;

                /*
                 * Directionality classify the difference block.
                 */
                diffDirIn = UCHAR_MAX;
                blockEdge(diffBlk, blkSizex, blkSizey, &diffDirIn);

                /*
                 * Adhoc lagrangian threshold used during
                 * search for best coding method. This threshold
                 * is dependent on whether the difference
                 * block is a smooth texture, coarse texture or
                 * some directionality or flat texture.
                 * For each such category a certain minimum bits
                 * and an acceptable square error define the
                 * minimum lagrangian achievable in that category.
                 */
                if(diffDirIn >= N_0_8 && diffDirIn <= N_7_8)
                  tmpThres = 5*lambda + 150;        /* 5, 100  */
                else if(diffDirIn == FLAT0)
                  tmpThres = 5*lambda + 150;        /* 5, 100  */
                else if(diffDirIn == TEXT1)
                  tmpThres = 6*lambda + 300;        /* 6, 150  */
                else if(diffDirIn == TEXT2)
                  tmpThres = 6*lambda + 450;        /* 6, 200 */
                else {
                  printf("Wrong diffDirectionality %d in intra\n", diffDirIn);
                  exit(1);
                }

                /*
                 * Test whether minimum achievable lagrangian
                 * can beat the not coded option. If it
                 * does then it is worth trying to encode
                 * the difference block.
                 */
                if(seBefore >= tmpThres)
                  codMeth8x8 = enc8x8Block(diffBlk, methCoeffs, qp,
                                           lambda, methods, class8x8,
                                           &cbsBits, &dctBits, &mesBits,
                                           &seAfter, predBlk, dctCoeffs,
                                           chroma, mode, predDirIn,
                                           1, prevLagr, cncDif,
                                           0, complexityMode);
                else
                  codMeth8x8 = 0;

                codMeth = codMeth8x8;
                coded   = (codMeth8x8 > 0) ? 1 : 0;
              }
              else if(blkSize == SIZE_4x4)
              {
                mesBits = 0;
                /*
                 * Collect the 12 pixel values of those prediction
                 * pixels which are adjacent to the border of the 4x4 block.
                 * They will be used for 4x4 extrapolation when
                 * calling enc4x4Block().
                 */
                pels    = getExtSouJunct(predBlk, ExtBlk, blkSizex, blkSizey);
                /*
                 * Try 4x4 coding of intra prediction error block.
                 */
                cbsBits = enc4x4Block(diffBlk, predBlk, origBlk, &seAfter,
                                      &dctBits, methCoeffs, 0, lambda, qp,
                                      dctCoeffs, ExtBlk, pels, chroma, mode,
                                      prevLagr, blkSize, 1, methName, cncDif,
                                      seBefore, 0, complexityMode);

                /*
                 * If we got a valid 4x4 method and lagrangian
                 * beats the not coded option then current block is coded.
                 * Otherwise the block is declared uncoded.
                 */
                if(methCoeffs[10] < AMOUNT_OF_4x4_METH &&
                   seBefore > seAfter + lambda * (cbsBits + dctBits))
                {
                  coded      = 1;
                  class4x4   = methCoeffs[13];
                  codMeth4x4 = methCoeffs[10];
                  codMeth8x8 = METH_4x4_1;
                }
                else
                {
                  coded      = 0;
                  class4x4   = 0;
                  codMeth4x4 = NOTCODED_METH;
                  codMeth8x8 = NOT_CODED;
                  seAfter    = seBefore;
                  cbsBits    = 0;
                  dctBits    = 0;
                }

                codMeth = codMeth4x4;
              }

              if(coded)
                comBit = codCom;
              else
              {
                comBit  = ncCom;
                seAfter = seBefore;
                cbsBits = 0;
                dctBits = 0;
                mesBits = 0;
              }
            }
            else /* Declare not coded */
            {
              coded   = 0;
              comBit  = ncCom;
              seAfter = seBefore;
              cbsBits = 0;
              dctBits = 0;
              mesBits = 0;
              codMeth = 0;
            }

            acBit = cbsBits + dctBits + mesBits;
            bits  = comBit + motBit + acBit;
            lagr  = seAfter + lambda * bits;

            /*
             * Final test to decide whether the
             * current intra prediction method
             * is the optimal so far.
             */
            if(lagr < bestLagr)
            {
              bestLagr    = lagr;
              bestMeth    = methName;
              bestDiff    = dcdIndex;
              bestCode    = coded;
              bestCom     = comBit;
              bestMot     = motBit;
              bestBmv     = motNo;
              bestValid   = noOfValidVectors;
              bestCnc     = cncDif;
              bestNo      = methRank;
              bestCM      = codMeth;
              bestPredDir = predDirIn;
              bestDiffDir = diffDirIn;
              bestEnergy  = seBefore;
              for(i=0; i<blkSize; i++)
              {
                tmpPredBlk[i] = predBlk[i];
                tmpDiffBlk[i] = diffBlk[i];
              }
            }

          }
        }
      }
    }

  }

  /*
   * As the best intra prediction method has been
   * selected a final full search for the prediction
   * error coding method is performed.
   */
  methName = bestMeth;
  class8x8 = methName + (NON_CLASSIFIED+1);
  blkClass = class8x8;
  for(i=0; i<blkSize; i++)
  {
    predBlk[i] = tmpPredBlk[i];
    diffBlk[i] = tmpDiffBlk[i];
  }

  seBefore = bestEnergy;

  coded   = 0;
  seAfter = seBefore;
  cbsBits = 0;
  dctBits = 0;
  mesBits = 0;
  /*
   * A test for fast estimation whether coding of
   * prediction error helps against not coded option.
   * In other words, if the not coded option has a
   * reasonable level of energy error compared to the
   * smallest cost of doing prediction error coding
   * then there is no need to further try coding the
   * prediction error.
   */
  if(seBefore > MIN_PRED_ERR_BITS * lambda + blkSize * ACCEPTABLE_MSE)
  {
    if(blkSize == SIZE_8x8)
    {
      /*
       * If the directionality of prediction block
       * has not yet been calculated then do it here.
       * Otherwise take the directionality of the
       * winning intra prediction method.
       */
      if(bestPredDir < 0)
      {
        if(methName == IP_METH_BM)
          blockEdge(predBlk, blkSizex, blkSizey, &predDirIn);
        else
          predDirIn = TEXT2;
      }
      else
        predDirIn = bestPredDir;

      /*
       * Do the actual full search for the best coding method
       * for the current 8x8 intra prediction error block.
       */
      codMeth8x8 = enc8x8Block(diffBlk, methCoeffs, qp, lambda, methods,
                               class8x8, &cbsBits, &dctBits, &mesBits,
                               &seAfter, predBlk, dctCoeffs, chroma, mode,
                               predDirIn, 0, HUGE_LAGRANGIAN_COST,
                               bestCnc, 0, complexityMode);
      if(codMeth8x8 == -1) {
        printf("Faulty condition in intra prediction\n");
        exit(1);
      }

      /*
       * Retain the best coding method and its class.
       */
      codMeth  = codMeth8x8;
      blkClass = class8x8;
    }
    else if(blkSize == SIZE_4x4)
    {
      /*
       * Collect the 12 pixel values of those prediction
       * pixels which are adjacent to the border of the 4x4 block.
       * They will be used for 4x4 extrapolation when
       * calling enc4x4Block().
       */
      pels    = getExtSouJunct(predBlk, ExtBlk, blkSizex, blkSizey);

      /*
       * Do the actual full search for the best coding method
       * for the current 4x4 intra prediction error block.
       */
      mesBits = 0;
      cbsBits = enc4x4Block(diffBlk, predBlk, origBlk, &seAfter, &dctBits,
                            methCoeffs, 0, lambda, qp, dctCoeffs,
                            ExtBlk, pels, chroma, mode, HUGE_LAGRANGIAN_COST, blkSize,
                            0, methName, bestCnc, seBefore,0, complexityMode);

      /*
       * The best 4x4 coding method is returned in methCoeffs[10]
       */
      if(methCoeffs[10] >= AMOUNT_OF_4x4_METH)
      {
        seAfter = seBefore;
        cbsBits = 0;
        dctBits = 0;
      }

      /*
       * One final check whether the best coded option can beat the
       * not coded alternative. Otherwise, declare the error block as not coded.
       */
      if(seBefore > seAfter + lambda*(cbsBits + dctBits))
      {
        class4x4   = methCoeffs[13];
        codMeth4x4 = methCoeffs[10];
        codMeth8x8 = METH_4x4_1;
      }
      else
      {
        class4x4   = 0;
        codMeth4x4 = NOTCODED_METH;
        codMeth8x8 = NOT_CODED;
        seAfter    = seBefore;
        cbsBits    = 0;
        dctBits    = 0;
      }

      /*
       * Retain the best coding method and its class.
       */
      blkClass = class4x4;
      codMeth  = codMeth4x4;
    }
    else {
      printf("Wrong blkSize %d in intra coding\n", blkSize);
      exit(1);
    }

    if(codMeth8x8 > 0)
      coded = 1;
    else
    {
      coded = 0;
      seAfter = seBefore;
      cbsBits = 0;
      dctBits = 0;
      mesBits = 0;
    }
  }

  /*
   * Inform about the bits for the chosen coding method.
   */
  acBit = cbsBits + dctBits + mesBits;

  if(coded != bestCode)
    if(coded > 0)
      bestCom += bestCnc;
    else
      bestCom -= bestCnc;

  bestCode = coded;
  bestAc   = acBit;
  bestErro = seAfter;
  bestLagr = bestErro + lambda*(bestCom+bestMot+bestAc);

  /*
   * In case the COEFF LAYER exists produce the
   * recosntructed block by adding on top of intra
   * prediction + dc difference the decoded
   * prediction error. Otherwise, the recosntructed
   * block is the same as the prediction + dcd.
   */
  if(bestCode > 0)
  {
    if(blkSize == SIZE_8x8)
      decBits = dec8x8(predBlk, diffBlk, methCoeffs, dctCoeffs, qp,
                       codMeth8x8, class8x8, chroma, mode, predDirIn);
    else if(blkSize == SIZE_4x4)
      decBits = dec4x4(predBlk, diffBlk, methCoeffs, dctCoeffs, qp,
                       ExtBlk, pels, chroma, mode, blkSize);
    else {
      printf("Wrong blkSize %d in YUV()\n", blkSize);
      exit(-1);
    }

    /*
     * Saturate to 0..255 range the reconstructed block.
     */
    for(i=0; i<blkSize; i++)
    {
      recoBlk[i] = predBlk[i] + diffBlk[i];
      if(recoBlk[i] > 255)    recoBlk[i] = 255;
      else if(recoBlk[i] < 0) recoBlk[i] = 0;
    }
  }
  else
    for(i=0; i<blkSize; i++)
      recoBlk[i] = predBlk[i];

  /*
   * Various consistency checks:
   */
  if(bestMeth < 0) {
    fprintf(stdout, "No intra prediction with predNo %d\n", predNo);
    exit(1);
  }
  if(bestDiff < -300) {
    printf("Error. Wrong bestDiff %d\n", bestDiff);
    exit(1);
  }
  if(bestCom < 0) {
    printf("Error. Wrong bestCom %d\n", bestCom);
    exit(1);
  }
  if(bestMot < 0) {
    printf("Error. Wrong bestMot %d\n", bestMot);
    exit(1);
  }
  if(bestNo < 0) {
    printf("Error. Wrong bestNo %d\n", bestNo);
    exit(1);
  }

  /*
   * Inform to upper level the bits and error spent
   * to intra encode the current block.
   */
  *acBits  = bestAc;
  *ovhBits = bestCom + bestMot;

  /*
   * Put the reconstructed block into the current
   * reconstruction frame buffer.
   */
  putIntBlktoByteImage(recoBlk, recoFr, blkSizex, blkSizey,
                       frX, frY, frSizex);

  /*
   * Next, the buffers classifBuf[] and dcFrameBuf[] are
   * updated to reflect the directionality and DC level
   * of the current reconstructed block. This information
   * is used when intra predicting the following intracoded
   * blocks inside the frame.
   * The "granularity" for luma is 8x8 whereas for chroma
   * is 4x4 irrespective if the current chroma block is 8x8.
   * Therefore, for chroma 8x8 the 4x4 left and 4x4 upper
   * neighbors of first quadrant are actually used for
   * intra DC prediction and context class building.
   */
  for(addy = 0; addy < blkSizey; addy += frStepy)
    for(addx = 0; addx < blkSizex; addx += frStepx)
    {
      addNo = addy / frStepy * xBlocks + addx / frStepx;

      /*
       * Extract the current block.
       */
      getIntBlkFromByteImage(recoFr, tmpBlk, frStepx, frStepy,
                             frX+addx, frY+addy, frSizex);

      /*
       * Directionality classify the block.
       */
      blockEdge(tmpBlk, frStepx, frStepy, &classifBuf[frBlkNo+addNo]);

      /*
       * Map the Texture coarse to texture smooth since
       * they are treated as same during context class building.
       */
      if(classifBuf[frBlkNo+addNo] == TEXT2)
        classifBuf[frBlkNo+addNo] = TEXT1;

      /*
       * Store also the quantised DC used later for intra DC prediction.
       */
      for(i=0, dcRecoSum = 0; i < frStepx * frStepy; i++)
        dcRecoSum += tmpBlk[i];

      dcFrameBuf[frBlkNo+addNo] = ipdDCquantAndInvQuant(dcRecoSum, frStepx*frStepy,
                                                        maxValue, maxLevel, &tmpInd);
    }

  if(gen_bitstr_flag)
  {
    /*
     * Here the actual dumping of bits to bitstream
     * takes place for this intra 8x8 or 4x4 block.
     */

    /*
     * Dump first the IPMR (intra prediction method rank),
     * then ICNCB (intra coded not coded block),
     * then IDCD (DC difference data, including IDCDSIGN bit).
     */
    bits = SendSelDcdCncSign_Event(predNo, bestNo, abs(bestDiff), bestCode,
                                   ((bestDiff < 0) ? -1 : 1), qp, 1);
    if (bestCom != bits) {
      fprintf(stderr, "Bits mismatch in intra pred methods dumping\n");
      exit(-1);
    }

    /*
     * In case of BM (intra pred method P13) dump IBMV
     * (intra block matching vector) data.
     */
    if (bestMeth == IP_METH_BM)
    {
      bits = SendBMV_Event(validY[bestBmv], validX[bestBmv], reducedSearchRange, 1);

      if (bestMot != bits) {
        fprintf(stderr, "Bits mismatch in intra BMV dumping\n");
        exit(-1);
      }
    }

    /*
     * In case the COEFF LAYER is present for the current block
     * then dump additional bits corresponding to intra prediction
     * error coding of this block.
     */
    if(bestCode > 0)
    {
      /*
       * These counters are used in the deeper functions
       * which store the TCOEFF data.
       */
      int
        tmpDouCnt = 0,
        tmpRdcCnt = 0,
        tmpVqCnt  = 0,
        tmpTriCnt = 0;

      doublet_counter = 0;
      rdc_counter     = 0;
      vq_counter      = 0;
      triplet_counter = 0;

      /*
       * Depending on block size first store the
       * given data in internal buffers and then
       * call dump8x8BlockAcInfo() or dump4x4BlockAcInfo()
       * to translate into bits and then dump to bitstream.
       */
      if(blkSize == SIZE_8x8)
      {

        /*
         * Determine how to translate a block having a certain
         * directionality into a tag specific to a set of
         * coding methods.
         */
        classTmp = changeClassNew(class8x8, predDirIn, qp, chroma,
                                  mode, &rotateFlag);

        /*
         * Store the 8x8 method name and the tag of
         * list from where it will be signalled as METH8
         */
        store_8x8_info1(codMeth8x8, classTmp, dctCoeffs, mode);

        /*
         * If we deal with a coding method having active
         * quadrants then store the METH4 - TCOEFF layer
         * for each of the active quadrant.
         */
        if(codMeth8x8 == METH_4x4_15)
        {
          /*
           * Store data related to signalling METH4.
           */
          store_8x8_info2(methCoeffs, codMeth8x8);
          dct4x4Meth = findDctMeth(methCoeffs, codMeth8x8, NULL);
          /*
           * Store data for TCOEFF in case we have a 4x4
           * transform for some active quadrants.
           */
          if(dct4x4Meth >= METH_4x4_DCT_1 && dct4x4Meth <= METH_4x4_DCT_15)
          {
            store_8x8_info3(dctCoeffs, dct4x4Meth, chroma, mode);
          }
        }

        /*
         * Store TCOEFF related data in case we deal with transform coding.
         */
        if((codMeth8x8 >= METH_8x8_DCT &&
            codMeth8x8 <= METH_8x3_DIAG_DCT) ||
           (codMeth8x8 >= METH_KLT_8x8_N_0_8 &&
            codMeth8x8 <= METH_KLT_8x8_TEXT) ||
           (codMeth8x8 >= METH_DCT_8x8_N_0_8 &&
            codMeth8x8 <= METH_DCT_8x8_N_4_8)) {
          store_8x8_info3(methCoeffs, codMeth8x8, chroma, mode);
        }

        /*
         * Actual dumping to bitstream based on the
         * information stored just above.
         */
        bits = dump8x8BlockAcInfo(mode, chroma, qp, &tmpDouCnt,
                                  &tmpRdcCnt, &tmpVqCnt, &tmpTriCnt, 1);
      }
      /*
       * Chroma of an 8x8 intra block in P frames
       * has always the size 4x4 and its coding info
       * is stored and dumped here.
       */
      else if(blkSize == SIZE_4x4)
      {
        /*
         * Store info related to METH4.
         */
        codMeth8x8 = METH_4x4_1;
        store_8x8_info2(methCoeffs, codMeth8x8);
        dct4x4Meth = findDctMeth(methCoeffs, codMeth8x8, NULL);

        /*
         * Store info related to TCOEFF if we are dealing
         * with 4x4 transform coding.
         */
        if(dct4x4Meth >= METH_4x4_DCT_1 &&
           dct4x4Meth <= METH_4x4_DCT_15) {
          for(i=16; i<64; i++)                /* Needed for 4x4 init */
            dctCoeffs[i] = 0;
          store_8x8_info3(dctCoeffs, dct4x4Meth, chroma, mode);
        }

        /*
         * Actual dumping to bitstream based on the
         * information stored just above.
         */
        bits = dump4x4BlockAcInfo(mode, chroma, qp, &tmpRdcCnt, &tmpVqCnt,
                                  &tmpTriCnt, 1);
      }

      if (bestAc != bits) {
        fprintf(stderr, "Bits mismatch in intra dumping\n");
        exit(-1);
      }

      /*
       * Verification of internal operation related to
       * storing and dumping of intra coding information.
       */
      if(tmpTriCnt != triplet_counter ||
         tmpDouCnt != doublet_counter ||
         tmpVqCnt  != vq_counter ||
         tmpRdcCnt != rdc_counter) {
        fprintf(stderr, "Counters mismatch in intra coding\n");
        exit(1);
      }

    }

  }

  /*
   * The embedded deblocking filtering is done within the intra block
   * coding loop just after the current block has been reconstructed.
   * The filtering is done in-place and for all picture components
   * (luminance, chorminance u and chrominance v) but only in I-frames.
   * For intra blocks in P-frames the filtering is done at the end
   * of frame coding in same manner as for inter blocks.
   */
  if(mode == FIRST_BLOCK)
    filCompFiltIntra(recoFr, frSizex, frSizey, qp, recoFr, frX, frY);
}


/***************************************************************************
 * Function    : SendIntraMccRegion()
 *
 * Purpose     : Encodes and dumps the MB/Block layer bitstream for Intra
 *               mode blocks or MBs. Intra prediction is used followed by
 *               MCC prediction error coding.
 *
 * Arguments   : See the function call
 *
 * Returns     : void
 *
 *               Copyright (c) 1997 Nokia Research Center
 ***************************************************************************/

/* qpBlockLuma used to pass qpY to vdxGetDctTable() and vdxGetKltTable() */
extern int
  qpBlockLuma;

void SendIntraMccRegion(struct frame *currFr, /* In:  Current (original) frame  */
                        struct frame *recoFr, /* Out: Reconstruction frame      */
                        int16 *slicePredMap,  /* In:  Slice label buffer        */
                        int currSliceID,      /* In:  Label of current slice    */
                        int sliceXmin,        /* In:  Xmin  of current slice    */
                        int sliceXmax,        /* In:  Xmax  of current slice    */
                        int sliceYmin,        /* In:  Ymin  of current slice    */
                        int sliceYmax,        /* In:  Ymax  of current slice    */
                        object *currBlockOrMBstruct, /* In: Ptr to MB/B object  */
                        int regionSize,       /* In:  Region size (16 or 8)     */
                        int predPossible,     /* In:  Is motion pred possible   */
                        int picSizeX,         /* In:  Width of the image        */
                        int picSizeY,         /* In:  Height of the image       */
                        int mode,             /* In:  Intra in I or P frames    */
                        int qp,               /* In:  Quant. param. for luma    */
                        float lambda,         /* In:  Lambda param. for luma    */
                        int *bitsPErrLuma,    /* Out: Bits for pred err in luma */
                        int *bitsPErrChroma,  /* Out: Bits for pred err chroma  */
                        int *bitsIPredLuma,   /* Out: Bits for luma intra pred  */
                        int *bitsIPredChroma, /* Out: Bits for chroma intra pred*/
                        int *bitsMode,        /* Out: Bits for mode,I frame only*/
                        int gen_bitstr_flag,  /* In:  Dump output to bitstream  */
                        int complexityMode)   /* In:  Level of encod complexity */
{
  Byte
    /*
     * Each input and output buffer has three components:
     * 0 (luma), 1 (chroma U), 2 (chroma V) which are
     * initialized in the followings.
     */
    *orig[3],
    *reco[3];

  static int16
    /*
     * Memory allocation for certain buffers are made
     * during the first invocation only.
     */
    first_invo = 1,
    /*
     * The buffer dcFrameBuf[0,1,2] holds the DC values for
     * all the luma and chroma blocks which have been so far
     * encoded. These values are used in DC intra prediction.
     * The allocation is made for every 8x8 luma and every
     * 4x4 chroma u and v blocks.
     */
    *dcFrameBuf[3];
  
  static u_char
    /*
     * The buffer classifBuf[0,1,2] holds the results of
     * classification for all the luma and chroma blocks
     * which have been so far encoded. This information is
     * used in context classification of Left and Upper
     * blocks. The allocation is made for every 8x8 luma
     * and every 4x4 chroma u and v blocks.
     */
    *classifBuf[3];

  int
    x, y, comp,
    sliceSx, sliceSy,
    xmin, xmax, ymin, ymax,
    chroma, xBlocks, yBlocks, qpY, qpUv, lambdaY, lambdaUv,
    frQp, frLambda, frSizex, frSizey, blkSizex, blkSizey,
    sliceMinx, sliceMaxx, sliceMiny, sliceMaxy,
    frStepx, frStepy, acBits, ovhBits,
    /*
     * Counters for number of bits used for
     * intra prediction for luma, chroma u,
     * and chroma v.
     */
    ipMethBits[3] = {0, 0, 0},
    /*
     * Counters for number of bits used for
     * intra prediction error coding for
     * luma, chroma u, and chroma v.
     */
    ipPErrBits[3]  = {0, 0, 0};

  orig[0] = currFr->yptr, orig[1] = currFr->uptr, orig[2] = currFr->vptr;
  reco[0] = recoFr->yptr, reco[1] = recoFr->uptr, reco[2] = recoFr->vptr;

    /*
   * Number of 8x8 blocks for horizontal
   * and vertical dimensions.
   */
  xBlocks = picSizeX/8;
  yBlocks = picSizeY/8;

  /*
   * Number of 8x8 blocks on horizontal
   * and vertical dimensions of the current
   * intra slice.
   */
  sliceSx = (sliceXmax-sliceXmin+1)/8;
  sliceSy = (sliceYmax-sliceYmin+1)/8;

  /*
   * QP and lambda used for luma and chroma.
   */
  /* qpBlockLuma used to pass qpY to vdxGetDctTable() and vdxGetKltTable() */
  qpBlockLuma = qpY = (int)qp;
  lambdaY  = (int)(lambda + 0.5);
  qpUv     = annexTlookup[qp];
  lambdaUv = (int) (0.85 * qpUv * qpUv);
  if (lambdaUv < 1)
    lambdaUv = 1;


  if (first_invo) 
  {
    first_invo = 0;
    /*
     * The buffer dcFrameBuf[0,1,2] holds the DC values for
     * all the luma and chroma blocks which have been so far
     * encoded. These values are used in DC intra prediction.
     * The allocation is made for every 8x8 luma and every
     * 4x4 chroma u and v blocks.
     */
    if ((dcFrameBuf[0] =
        (int16 *)malloc(3 * xBlocks * yBlocks * sizeof(int16))) 
        == NULL)
    {
      deb0f(stderr,"Cannot allocate static memory for dcFrameBuf\n");
      nccExit(1);
    }
    dcFrameBuf[1] = dcFrameBuf[0] + xBlocks * yBlocks;
    dcFrameBuf[2] = dcFrameBuf[0] + (xBlocks * yBlocks << 1);

    /*
     * The buffer classifBuf[0,1,2] holds the results of
     * classification for all the luma and chroma blocks
     * which have been so far encoded. This information is
     * used in context classification of Left and Upper
     * blocks. The allocation is made for every 8x8 luma
     * and every 4x4 chroma u and v blocks.
     */
    if((classifBuf[0] =
        (u_char *)malloc(3 * xBlocks * yBlocks * sizeof(u_char))) 
       == NULL) 
    {
      deb0f(stderr,"Cannot allocate static memory for classifBuf\n");
      nccExit(1);
    }
    classifBuf[1] = classifBuf[0] + xBlocks * yBlocks;
    classifBuf[2] = classifBuf[0] + (xBlocks * yBlocks << 1);
  }

  /*
   * Send Mode information for intra macroblocks or
   * 8x8 blocks inside P-frames only. For I-frames
   * there is no mode information to be sent.
   * The current region can be either an 16x16 macroblock
   * (four 8x8 luma + two 8x8 chroma) or 8x8 block
   * (one 8x8 luma + two 4x4 chroma).
   */

  if (mode == INTRA_BLOCK)
    *bitsMode = SendRegionMode(REGMODE_INTRA, 0 /*dummy*/,
                               predPossible, regionSize,
                               gen_bitstr_flag);
  else
    *bitsMode = 0;


  /*
   * For each frame component (0 = luma, 1 = chroma u, 2 = chroma v)
   * do the intra encoding and sending of bits for the 8x8 or 4x4
   * blocks contained within the MB or block. This order is important
   * because sending of bits takes place inside codeIntraBlock()
   * function.
   */
  for (comp = 0; comp < 3; comp++)
  {
    switch (comp)
    {
      /* Luminance parameters for comp 0 (luma) */
      case 0:

        chroma    = LUMA;
        frQp      = qpY;
        frLambda  = lambdaY;
        blkSizex  = blkSizey = 8;
        frStepx   = frStepy  = 8;
        frSizex   = picSizeX;
        frSizey   = picSizeY;
        sliceMinx = (sliceXmin+0);
        sliceMaxx = (sliceXmax+1);
        sliceMiny = (sliceYmin+0);
        sliceMaxy = (sliceYmax+1);
        xmin      = currBlockOrMBstruct->xmin;
        xmax      = currBlockOrMBstruct->xmax;
        ymin      = currBlockOrMBstruct->ymin;
        ymax      = currBlockOrMBstruct->ymax;

        break;

      /*
       * Chrominance parameters for comp 1 (chroma U)
       * are chronologically set after the luminance
       * component is completed.
       */
      case 1:

        chroma    = CHROMA;
        frQp      = qpUv;
        frLambda  = lambdaUv;
        blkSizex  = blkSizey = ((regionSize == 16) ? 8 : 4);
        frStepx   = frStepy  = 4;
        frSizex   /= 2;
        frSizey   /= 2;
        sliceMinx /= 2;
        sliceMaxx /= 2;
        sliceMiny /= 2;
        sliceMaxy /= 2;
        xmin      /= 2;
        xmax      /= 2;
        ymin      /= 2;
        ymax      /= 2;

        break;

      /*
       * Chrominance parameters for comp 2 (chroma V)
       * are the same as for chroma U component.
       */
      case 2:

        break;

    }

    for(y = ymin; y < ymax; y += 8)
    {
      for(x = xmin; x < xmax; x += 8) 
      {
        /*
         * Intra encode and send the current 8x8 or 4x4 block
         * inside the given luma or chroma component.
         */
        codeIntraBlock(orig[comp], reco[comp],
                       dcFrameBuf[comp], classifBuf[comp],
                       &acBits, &ovhBits,
                       mode, chroma, frQp, frLambda,
                       maxDcLevels[frQp],
                       blkSizex, blkSizey, x, y, frStepx, frStepy,
                       frSizex, frSizey, xBlocks,
                       gen_bitstr_flag, sliceMinx, sliceMaxx,
                       sliceMiny, sliceMaxy, slicePredMap,
                       sliceSx, sliceSy, currSliceID, complexityMode);

        /*
         * Accumulate the amount of bits for
         * IPMR, ICNCB, IDCD and IBMV.
         */
        ipMethBits[comp] += ovhBits;

        /*
         * Accumulate the amount of bits for
         * COEFF LAYER.
         */
        ipPErrBits[comp] += acBits;

      }
    }
  }

  /*
   * Total amount of bits for the current
   * region. The current region can be either
   * an 16x16 macroblock (four 8x8 luma + two 8x8 chroma)
   * or 8x8 block (one 8x8 luma + two 4x4 chroma).
   */
  *bitsIPredLuma   = ipMethBits[0];
  *bitsIPredChroma = ipMethBits[1] + ipMethBits[2];
  *bitsPErrLuma    = ipPErrBits[0];
  *bitsPErrChroma  = ipPErrBits[1] + ipPErrBits[2];

  /*
   * Next P region or next P-frame requires these
   * counters to be initialized beforehand.
   */
  if(gen_bitstr_flag)
  {
    doublet_counter = 0;
    triplet_counter = 0;
    vq_counter      = 0;
    rdc_counter     = 0;
  }

}


/***************************************************************************
 * Function    : Code_I_Frame()
 *
 * Purpose     : Encodes and dumps the frame & slice layer bitstream for Intra
 *               coded pictures. Intra prediction is used followed by
 *               MCC prediction error coding.
 *
 * Arguments   : See the function call
 *
 * Returns     : void
 *
 *               Copyright (c) 1997 Nokia Research Center
 ***************************************************************************/

int Code_I_Frame(struct frame *origFr, /* In:  Current (original) frame      */
                 struct frame *recoFr, /* Out: Reconstruction frame          */
                 int picSizeX,         /* In:  Width of the image            */
                 int picSizeY,         /* In:  Height of the image           */
                 int frameNumber,      /* In:  Picture frame number          */
                 int qp,               /* In:  Quant. param. for luma        */
                 Bits *bits,           /* Out: Picture amount of bits        */
                 int16 *slicePredMap,  /* In:  Slice label buffer            */
                 int sliceMode,        /* In:  Flag to signal slice mode     */
                 int sliceBits,        /* In:  Recommended slice size in bits*/
                 int complexityMode)   /* In:  Level of encoder complexity   */
{
  /*
   * Lambda parameter for I-frames is slightly
   * lower than the corresponding one for P-frames.
   */
  float
    lambda = (float) (0.7 * qp * qp);


  /*
   * The current Macroblock has a corresponding
   * data structure holding geometry and other
   * type of information.
   */
  object
    currMBstruct;

  int
    x, y, mbNum, i,
    picSize = picSizeX * picSizeY,
    currSliceID = 0, sliceCurrBits, frameBitsAtStartSlice = 0,
    sliceYmin, sliceXmin, sliceYmax, sliceXmax,
    bitsPErrLuma, bitsPErrChroma, bitsIPredLuma, bitsIPredChroma, bitsMode;

  /*
   * Frame number and picture type echo.
   */
  fprintf(stderr, "Coding I-Frame %d\n", frameNumber);

  /*
   * Bitstream internal counters are reset at beginning
   * a frame coding.
   */
  reset_all_bits();
  /*
   * Fields inside "bits" structure are made zero.
   */
  ResetBitCount(bits);

  /*
   * The slicePredMap tells encoder the causal zone from where it
   * can use intra prediction to predict the current intra macroblock.
   * Slice parameters and slice map are initialized
   * before the coding loop starts.
   */
  for (i = (picSize >> 6) - 1; i >= 0; i--)
    slicePredMap[i] = SHRT_MAX;

  sliceYmin = picSizeY;
  sliceXmin = picSizeX;
  sliceYmax = 0;
  sliceXmax = 0;

  /*
   * Internal encoder buffers used in dumping of
   * bits for prediction error coding are allocated here
   * if they are not already allocated.
   */
  alloc_triplets_if_needed(picSize);

  /*
   * Send I-frame header to bitstream.
   */
  bits->header = SendPicHead(2, frameNumber, sliceMode, 1,
                             qp, picSizeX, picSizeY,
                             0, NULL);
  if (sliceMode)
  {

    /*
     * Dump the header of very first Slice's without an SSC
     */
    bits->header += SendSliceHead(currSliceID, 1, qp, frameNumber,
                                  picSizeX, picSizeY, 0,
                                  0, 2);

    /*
     * Echo coords of the slice.
     */
    fprintf(stderr, "First slice start at I %3d %2d\n",
            frameNumber, currSliceID);
  }


  /*
   * Macroblock intra coding loop starts.
   */

  for (y=0, mbNum=0; y<picSizeY; y += 16)
    for (x=0; x<picSizeX; x += 16, mbNum++)
    {

      if (sliceMode && sliceBits)
      {
        /*
         * If slice mode is enabled and
         * a new slice is decided then the
         * slice ID will be the number of first MB
         * inside the slice.
         */
        sliceCurrBits = get_all_bits() - frameBitsAtStartSlice;

        if ((sliceBits < 0) ? (mbNum%((picSizeX/16)*(-sliceBits)) == 0 && mbNum!=0) :
                              (sliceCurrBits > sliceBits))
        {
          currSliceID = mbNum;

          /*
           * Stuffing the bitstream before the next slice SSC is dumped.
           */
          put_bits(0,0,1,1);

          frameBitsAtStartSlice = get_all_bits();

          /*
           * Dump header of a new slice (SSC+mbNum+etc) to the bitstream
           * and re-initialize slice geometry parameters.
           */

          bits->header += SendSliceHead(currSliceID, 1, qp, frameNumber,
                                        picSizeX, picSizeY, 1,
                                        1, 2);
          sliceYmin = picSizeY;
          sliceXmin = picSizeX;
          sliceYmax = 0;
          sliceXmax = 0;

          /*
           * Echo coords of the slice.
           */
          fprintf(stderr, "Slice start at I %3d %2d\n",
                  frameNumber, currSliceID);
        }
      }

      /*
       * The slicePredMap tells encoder the causal zone from where it
       * can use intra prediction to predict the current intra macroblock.
       * Here, all the four blocks of the current MB take the same
       * label as the current slice (which label is currSliceID).
       */
      i = (y/8)*(picSizeX/8) + (x/8);
      slicePredMap[i]               =
      slicePredMap[i+1]             =
      slicePredMap[i+picSizeX/8]    =
      slicePredMap[i+picSizeX/8+1]  = currSliceID;

      /*
       * Update the geometry of the slice which has developped so far.
       */
      if (y < sliceYmin) sliceYmin = y;
      if (x < sliceXmin) sliceXmin = x;
      if (y+15 > sliceYmax) sliceYmax = y+15;
      if (x+15 > sliceXmax) sliceXmax = x+15;

      /*
       * Current region is always a 16x16 pel macroblock.
       */
      currMBstruct.xmin  =     x;
      currMBstruct.xmax  =  x+15;
      currMBstruct.ymin  =     y;
      currMBstruct.ymax  =  y+15;

      /*
       * Actual intra coding of current MB (given its position
       * and the geometry of the slice it belongs to).
       */
      SendIntraMccRegion(origFr, recoFr, slicePredMap, currSliceID,
                         sliceXmin, sliceXmax, sliceYmin, sliceYmax,
                         &currMBstruct, 16, 0,
                         picSizeX, picSizeY, FIRST_BLOCK, qp, lambda,
                         &bitsPErrLuma, &bitsPErrChroma,
                         &bitsIPredLuma, &bitsIPredChroma,
                         &bitsMode, 1, complexityMode);

      /*
       * Frame bit counters are updated.
       */
      bits->coefy += bitsPErrLuma;
      bits->coefu += bitsPErrChroma;
      bits->cncy  += bitsIPredLuma;
      bits->cncuv += bitsIPredChroma;
      bits->mode  += bitsMode;
    }

  /*
   * Byte align followed by flushing of the bitstream.
   */
  put_bits(0,0,1,1);
  fflush(bitstream_bit);

  /*
   * I-frame total amount of bits is computed and returned.
   */
  bits->coefv = 0;
  bits->total = bits->coefy   + bits->coefu + bits->coefv +
                bits->cncy    + bits->cncuv + bits->mode  +
                bits->header;

  return bits->total;
}


/*
 *
 *  findIntraMotion:
 *
 * Parameters:
 *      origBlk         I: Original block.
 *      recoImag        I: Reconstructed Image.
 *      predBlk         O: Predicted block.
 *      x               I: x Position.
 *      y               I: y Position.
 *      motNo           I: Number of motion vector.
 *      blocksizex      I: width of the block.
 *      blocksizey      I: length of the block.
 *      x_size          I: width of the Image.
 *      y_size          I: length of the Image.
 *      prevMin         I: Value of previous best lagrangian.
 *      lambda          I: Value of lambda.
 *
 * Function:
 *      Simple pel block matching search for the best vector.
 *      This function is used in I frame intra prediction.
 *
 * Returns:
 *      dcValueSum      Sum of prediction block values.
 *
 * Note: This function is currently not used at runtime since BM in
 *       intra prediction is disabled.
 */
static int32 findIntraMotion(int32 *origBlk, u_char *recoImag, int32 *predBlk,
                             int32 x, int32 y, int32 *motNo, int32 blocksizex,
                             int32 blocksizey, int32 x_size, int32 prevMin,
                             int32 lambda, int32 *bits, int reducedSearchRange,
                             int32 noOfValidVectors, int32 *validX, int32 *validY)
{
  int
    pelNo, vectNo, dX, dY,
    dcValueSum, lagr, tmp, size,
    tmpBits, gen_bitstr_flag = 0,
    pelX, pelY;

  *motNo     = -1;
  *bits      = 50;
  dcValueSum = -1;

  size = blocksizex*blocksizey;
  for(vectNo=0; vectNo<noOfValidVectors; vectNo++) {
    dX = validX[vectNo];
    dY = validY[vectNo];

    tmpBits = SendBMV_Event(dY, dX, reducedSearchRange, 0);
    lagr    = lambda*tmpBits;

    pelNo = 0;
    pelY = 0;
    while(lagr<prevMin && pelY<blocksizey) {
      pelX = 0;
      while(lagr<prevMin && pelX<blocksizex) {
        tmp = origBlk[pelNo] - recoImag[(y+dY+pelY)*x_size + (x+dX+pelX)];
        lagr += tmp*tmp;
        pelNo++;
        pelX++;
      }
      pelY++;
    }

    /*
     * If the new best motion vector has found.
     */
    if(lagr < prevMin) {
      prevMin = lagr;
      *motNo  = vectNo;
      *bits   = tmpBits;

      dcValueSum = 0;
      pelNo = 0;
      pelY = 0;
      while(pelY < blocksizey) {
        pelX = 0;
        while(pelX < blocksizex) {
          tmp = recoImag[(y+dY+pelY)*x_size + (x+dX+pelX)];
          predBlk[pelNo] = tmp;
          dcValueSum+= tmp;
          pelNo++;
          pelX++;
        }
        pelY++;
      }
    }
  }

  return dcValueSum;
}


