#ifndef _TOOLMCC_H_
#define _TOOLMCC_H_

/*
 * toolMCC.h
 *
 *      Copyright (c) 1995-2000 Nokia Research Center
 *
 * Contains:
 *      Tool functions for multiple choice coding (MCC)
 *      MCC is used for prediction error coding
 *      File Contains also general tools
 *
 */


#include "mainFlags.h"


/*
 * Defines
 */

#define CALC_STAT           0
/*
 * Directions of edges in 4x4 or 8x8 block.
 */
#define N_0_8               0
#define N_1_8               1
#define N_2_8               2
#define N_3_8               3
#define N_4_8               4
#define N_5_8               5
#define N_6_8               6
#define N_7_8               7
#define FLAT0               8
#define TEXT1               9
#define TEXT2              10

/*
 * Different 8x8 prediction classes.
 */
#define QUAD_4_1           13   /* all quadrants                        */
#define NON_CLASSIFIED     25   /* intra or non classified inter        */

/*
 * Amount of pels in different block sizes.
 */
#define COLS_8x8            8
#define ROWS_8x8            8
#define SIZE_8x8            64
#define COLS_8x4            8
#define ROWS_8x4            4
#define SIZE_8x4            32
#define COLS_3x8            3
#define ROWS_3x8            8
#define SIZE_3x8            24
#define COLS_4x4            4
#define ROWS_4x4            4
#define SIZE_4x4            16
#define COLS_4x8            4
#define ROWS_4x8            8
#define SIZE_4x8            32

/*
 * Locations of 4x4 or 4x8 blocks.
 */
#define QUAD_1              1   /* top left quadrant            */
#define QUAD_2              2   /* top right quadrant           */
#define QUAD_3              3   /* bottom left quadrant         */
#define QUAD_4              4   /* bottom right quadrant        */
#define JUNCTION            5   /* 4x4 junction color block     */
#define HALF_1              6   /* top (or left) half           */
#define HALF_2              7   /* bottom (or right) half       */
#define HALF_3              8   /* left half                    */
#define HALF_4              9   /* right half                   */

/*
 * Starting rows, columns ans values for 8x4 block.
 */
#define START_COL_8x4_1     0
#define START_ROW_8x4_1     0
#define START_VAL_8x4_1     0
#define START_COL_8x4_2     0
#define START_ROW_8x4_2     ROWS_8x4
#define START_VAL_8x4_2     SIZE_8x4

/*
 * Starting rows, columns ans values for 4x8 block.
 */
#define START_COL_4x8_1     0
#define START_ROW_4x8_1     0
#define START_VAL_4x8_1     0
#define START_COL_4x8_2     COLS_4x8
#define START_ROW_4x8_2     0
#define START_VAL_4x8_2     SIZE_4x8
#define START_COL_4x8_4     2
#define START_ROW_4x8_4     0
#define START_VAL_4x8_4     0

/*
 * Starting rows, columns ans values for 4x4 block.
 */
#define START_COL_4x4_1     0
#define START_ROW_4x4_1     0
#define START_VAL_4x4_1     0
#define START_COL_4x4_2     COLS_4x4
#define START_ROW_4x4_2     0
#define START_VAL_4x4_2     SIZE_4x4
#define START_COL_4x4_3     0
#define START_ROW_4x4_3     ROWS_4x4
#define START_VAL_4x4_3     32
#define START_COL_4x4_4     COLS_4x4
#define START_ROW_4x4_4     ROWS_4x4
#define START_VAL_4x4_4     48


/*
 * Different 4x4 methods.
 */

#define NOTCODED_METH       0
#define EXTRAPOL_METH       1
#define REMOVE_DC_4x4       2
#define METH_DCT_4x4        3

#define METH_DCT_4x4_N_0_8  4
#define METH_DCT_4x4_N_1_8  5
#define METH_DCT_4x4_N_2_8  6
#define METH_DCT_4x4_N_3_8  7
#define METH_DCT_4x4_N_4_8  8

#define KLT_4x4_N_0_8      32   /* Not Used */
#define KLT_4x4_N_1_8      33
#define KLT_4x4_N_2_8      34
#define KLT_4x4_N_3_8      35
#define KLT_4x4_N_4_8      36   /* Not Used */
#define KLT_4x4_N_5_8      37
#define KLT_4x4_N_6_8      38
#define KLT_4x4_N_7_8      39
#define KLT_4x4_TEXT       40   /* Not Used */


/*
 * Different 8x8 methods.
 */
#define NOT_CODED           0
#define METH_8x8_DCT        1
#define METH_4x8_DCT_1      2   /* left half                            */
#define METH_4x8_DCT_2      3   /* right half                           */
#define METH_4x8_DCT_3      4   /* left and right halfs                 */
#define METH_4x8_DCT_4      5   /* middle 4x8                           */
#define METH_8x4_DCT_1      6   /* top half                             */
#define METH_8x4_DCT_2      7   /* bottom half                          */
#define METH_8x4_DCT_3      8   /* top and bottom halfs                 */
#define METH_8x4_DCT_4      9   /* middle 8x4                           */
#define METH_3x8_DIAG_DCT  10   /*      \                               */
#define METH_8x3_DIAG_DCT  11   /*      /                               */

#define METH_REMOVE_DC_8x8 13

#define METH_DCT_ADAPTIVE  14   /* Not Used */
#define METH_KLT_ADAPTIVE  15   /* Not Used */

#define METH_EXTRAPOL      12   /* Not Used */

#define METH_4x4_DCT_1     13   /* top left quadrant                    */
#define METH_4x4_DCT_2     14   /* top right quadrant                   */
#define METH_4x4_DCT_3     15   /* top left and top right quadrants     */ 
#define METH_4x4_DCT_4     16   /* bottom left quadrant                 */
#define METH_4x4_DCT_5     17   /* top left and bottom left quadrants   */
#define METH_4x4_DCT_6     18   /* top right and bottom left quadrants  */
#define METH_4x4_DCT_7     19   /* all but no bottom right quadrant     */
#define METH_4x4_DCT_8     20   /* bottom right quadrant                */
#define METH_4x4_DCT_9     21   /* top left and bottom right quadrants  */
#define METH_4x4_DCT_10    22   /* top right and bottom right quadrants */
#define METH_4x4_DCT_11    23   /* all but no bottom left quadrant      */
#define METH_4x4_DCT_12    24   /* bottom left and bottom right quads   */
#define METH_4x4_DCT_13    25   /* all but no top right quadrant        */
#define METH_4x4_DCT_14    26   /* all but no top left quadrant         */
#define METH_4x4_DCT_15    27   /* all quadrants                        */

#define METH_4x4_1         28   /* top left quadrant                    */
#define METH_4x4_2         29   /* top right quadrant                   */
#define METH_4x4_3         30   /* top left and top right quadrants     */
#define METH_4x4_4         31   /* bottom left quadrant                 */
#define METH_4x4_5         32   /* top left and bottom left quadrants   */
#define METH_4x4_6         33   /* top right and bottom left quadrants  */
#define METH_4x4_7         34   /* all but no bottom right quadrant     */
#define METH_4x4_8         35   /* bottom right quadrant                */
#define METH_4x4_9         36   /* top left and bottom right quadrants  */
#define METH_4x4_10        37   /* top right and bottom right quadrants */
#define METH_4x4_11        38   /* all but no bottom left quadrant      */
#define METH_4x4_12        39   /* bottom left and bottom right quads   */
#define METH_4x4_13        40   /* all but no top right quadrant        */
#define METH_4x4_14        41   /* all but no top left quadrant         */
#define METH_4x4_15        42   /* all quadrants                        */

#define METH_KLT_8x8_N_0_8 43   /* Not Used */
#define METH_KLT_8x8_N_1_8 44
#define METH_KLT_8x8_N_2_8 45
#define METH_KLT_8x8_N_3_8 46
#define METH_KLT_8x8_N_4_8 47   /* Not Used */
#define METH_KLT_8x8_N_5_8 48
#define METH_KLT_8x8_N_6_8 49
#define METH_KLT_8x8_N_7_8 50
#define METH_KLT_8x8_TEXT  51   /* Not Used */

#define METH_DCT_8x8_N_0_8 52
#define METH_DCT_8x8_N_1_8 53
#define METH_DCT_8x8_N_2_8 54
#define METH_DCT_8x8_N_3_8 55
#define METH_DCT_8x8_N_4_8 56

#define METH_EXTRAPOL_10   62
#define METH_EXTRAPOL_20   63
#define METH_EXTRAPOL_40   64
#define METH_EXTRAPOL_80   65
#define METH_EXTRAPOL_12   66
#define METH_EXTRAPOL_14   67
#define METH_EXTRAPOL_28   68
#define METH_EXTRAPOL_48   69


/*
 * For dcts VLC tables.
 */
#define LAST                     2
#define RUN                     64
#define VLC_TYPES                6

#define LEVEL_12              4096
#define LEVEL_11              2048
#define LEVEL_10              1024
#define LEVEL_09               512
#define LEVEL_08               256
#define LEVEL_07               128
#define LEVEL_06                64
#define LEVEL_05                32
#define LEVEL_04                16
#define LEVEL_03                 8
#define LEVEL_02                 4
#define LEVEL_01                 2


/*
 * Type of image block.
 */
#define LUMA                0
#define CHROMA              1
#define ESCAPE             -1   /* Escape code needed in vlc            */


#define AMO_RECO_TAB       16           /* 0 - 15 (level)         */
#define AMO_QP             32           /* 0 - 31 (qp)            */
#define MAX_RECO_TYPE      (4 * AMO_QP) /* XY, XC, IFY, IFC       */

#define AMOUNT_OF_HUFF     75
#define MAXLEN_OF_HUFF     12

#define DCT_TRANSFORM       0
#define KLT_TRANSFORM       1


/*
 * Amount of different 8x8 classes in overhead table, 
 */
#define AMOUNT_OF_8x8_CLASS_OVH   8

/* 
 * Offset to the start of CHROMA in overhead table, 
 * consist of INTER and INTRA of LUMA.
 */ 
#define START_OF_8x8_CHROMA_OVH  (4 * AMOUNT_OF_8x8_CLASS_OVH) 

#define AMOUNT_OF_8x8_OVH        (6 * AMOUNT_OF_8x8_CLASS_OVH)

#define AMOUNT_OF_8x8_METH       86 /* Amount of different 8x8 methods       */

#define MAX_OF_8x8_METH_OVH      12
#define AMOUNT_OF_KLT_8x8        2  /* Real amount of different klts        */
#define KLT_LEN_8x8              16  /* Maximum size of klt basisfunctions   */


/*
 * Amount of different 4x4 classes in overhead table, 
 */
#define AMOUNT_OF_4x4_CLASS_OVH    2

/* 
 * Offset to the start of CHROMA in overhead table, 
 * consist of INTER and INTRA of LUMA.
 */
#define START_OF_4x4_CHROMA_OVH    (4 * AMOUNT_OF_4x4_CLASS_OVH) 

#define AMOUNT_OF_4x4_OVH          (6 * AMOUNT_OF_4x4_CLASS_OVH)   

#define AMOUNT_OF_4x4_METH      60 /* Amount of different 4x4 methods       */

#define MAX_OF_4x4_METH_OVH     12
#define AMOUNT_OF_KLT_4x4       2  /* Amount of different klt:s            */
#define KLT_LEN_4x4             8  /* Maximum size of klt basisfunctions   */


#define EDGE_THR          1300  /* 4x4 classifiers variance threshold    */


/*
 * Operations needed to change the orientation
 * of block.
 */

#define UNIT_OPER        0
#define ROTATE_LEFT      1
#define ROTATE_RIGHT     2
#define ROTATE_180       3
#define MIRROR_VERT      4
#define MIRROR_HORI      5
#define MIRROR_DIAG      6

/*
 * Maximum amount of intra prediction methods
 * inside a context dependent list. 
 */
#define MAX_IP_LIST_SIZE   6


/*
 * Structs and typedefs
 */

extern int kltSize4x4[AMOUNT_OF_KLT_4x4]; 
extern int kltSize8x8[AMOUNT_OF_KLT_8x8]; 
extern u_char ipContextLists[49*(1+MAX_IP_LIST_SIZE)];
extern u_char overheadOrder8x8[], overheadOrder4x4[];
extern u_char overheadNumbers8x8[], overheadNumbers4x4[];

/*
 * Function prototypes
 */
int median(int *value, int size);

int ipdDCquantAndInvQuant(int cumDiff, int pels, int maxVal, int levels, int *index);

int ipdDCinvQuant(int maxVal, int levels, int level);

int predictDCValue(int left, int upper, int upperLeft, int chroma);

int bitsForDCDist(int distance, int chroma);

int getExtSouJunct(int *source, int *destination, int cols, int rows);

int getExtSou8x8(int *source, int *destination, int quadrant);

int findPredNum4x4(int *pred, int qp, int chroma, int mode, int blockSize,
                   int predDir, int *rotate);
int changeClassNew(int classIn, int predDirIn, int qp,
                   int chroma, int mode, int *rotateFlag);

void getBlock(int *source, int *destination, int rotateFlag);

void putBlock(int *source, int *destination, int rotateFlag);

int methSelectNum8x8New(int class, int method);

int methSelectNum4x4New(int class, int method);

int methSelectNumPreNew(int class, int method);

int newQuantCoef(double *dctCoef, int *quantDctCoef, int size, int qp,
                 int chroma, int mode, int transform);

void newInvQuantCoef(int *quantDctCoef, double *dctCoef, int size, int qp,
                     int chroma, int mode);

int MCCQuantCoef(int *dctCoef, int *quantDctCoef, int size, int qp,
                 int chroma, int mode, int transform);

int RD_Optimized_Quantize_Coef(int *dctCoef, int *quantDctCoef, int size, int qp, int chroma, 
                               int mode, int transform, u_char *zigZagTable, int sizeX, int sizeY, int kltNum);

void MCCInvQuantCoef(int *quantDctCoef, int *dctCoef, int size, int qp,
                     int chroma, int mode);

void KltTransform8x8(int *diff, int *coeffs, int32 kltNum, int32 mode);

void KltTransform4x4(int *diff, int *coeffs, int32 kltNum, int32 mode);

void KltInverseTransform8x8(int *coeffs, int *reco, int32 kltNum, int32 mode);

void KltInverseTransform4x4(int *coeffs, int *reco, int32 kltNum, int32 mode);

void varSizeDct(double *source, double *destination, int cols, int rows);

void varSizeInvDct(double *source, double *destination, int cols, int rows);

int findDctMeth(int *methInfo, int method8x8, int *save4);

int putIntBlock(int *block, int *image, int cols, int rows,
                int colStart, int rowStart, int colsIma);

int getIntBlock(int *image, int *block, int cols, int rows,
                int colStart, int rowStart, int colsIma);

int calcBlockEnergy(int *block, int size);

void findPredNums8x8(int *pred, int class, int qp, int *dctcoef, int chroma,
                     int method, int mode);

int detectClass8x8(int *source, float mismatchEps);

int getIntBlkFromByteImage(u_char *image, int *block, int cols, int rows,
                           int colStart, int rowStart, int colsIma);

int putIntBlktoByteImage(int *block, u_char *image, int cols, int rows,
                         int colStart, int rowStart, int colsIma);

void blockEdge(int *block, int sizex, int sizey, u_char *edge);

int32 predIntra(u_char *RecoImag, int32 *pred_blk, int32 x, int32 y,
                int32 method, int32 blocksizex, int32 blocksizey,
                int32 x_size, int32 y_size, int32 sliceMinx, int32 sliceMaxx,
                int32 sliceMiny, int32 sliceMaxy, int32 leftDc, int32 upperDc,
                int32 upleftDc, int16 *slicePredMap, int32 sliceSx,
                int32 sliceSy, int32 currSliceID);

int32 findValidVectors(int32 *validX, int32 *validY, int32 *validI,
                       int32 maxValid, int32 maxVector, int32 x, int32 y,
                       int32 blocksizex, int32 blocksizey, int32 x_size,
                       int32 sliceMinx, int32 sliceMaxx, int32 sliceMiny,
                       int32 sliceMaxy, int16 *slicePredMap,
                       int32 sliceSx, int32 sliceSy, int32 currSliceID,
                       int32 mode);

int32 predIntraBlockMatch(u_char *recoImag, int32 *predBlk, int32 x, int32 y,
                          int32 blocksizex, int32 blocksizey,
                          int32 x_size, int32 dY, int32 dX);

int find4x4Methods(int method);

int find4x4Start(int codingMethod, int size);

int findNumMethods(int method, int blkSize);

int find8x8Start(int codingMethod, int size);

void findTripletInfos(int method8x8, int method4x4, int qp, int chroma,
                      int mode, int *tableNo, int *runBits, int *levelBits);

void copyBlock4x4(int *source, int *destination, int rotateFlag);

void copyBlock8x8(int *source, int *destination, int rotateFlag);

#endif /* #ifndef _TOOLMCC_H_ */

