/*
 * toolMCCtable.c
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Contains:
 *    Look-up tables for prediction error coding, coding method signaling
 *    and intra prediction.
 */


#include "toolMCC.h"


/*
 * Index buffers consisting rotation or mirror information
 * for 4x4 block. These are used for 4x4 block manipulation.
 */

u_char
  rotateLeft4x4[16]  = { 3,  7, 11, 15,
                         2,  6, 10, 14,
                         1,  5,  9, 13,
                         0,  4,  8, 12},

  rotateRight4x4[16] = {12,  8,  4,  0,
                        13,  9,  5,  1,
                        14, 10,  6,  2,
                        15, 11,  7,  3},

  mirrorVert4x4[16]  = {12, 13, 14, 15,
                         8,  9, 10, 11,
                         4,  5,  6,  7,
                         0,  1,  2,  3},

  mirrorHori4x4[16]  = { 3,  2,  1,  0,
                         7,  6,  5,  4,
                        11, 10,  9,  8,
                        15, 14, 13, 12},

  mirrorDiag4x4[16]  = { 0,  4,  8, 12, 
                         1,  5,  9, 13,
                         2,  6, 10, 14,
                         3,  7, 11, 15},

  rotate180_4x4[16]  = {15, 14, 13, 12, 
                        11, 10,  9,  8,
                         7,  6,  5,  4,
                         3,  2,  1,  0};


/*
 * Index buffers consisting rotation or mirror information
 * for 8x8 block. These are used for 8x8 block manipulation.
 */

u_char
  rotateLeft8x8[64]  = { 7, 15, 23, 31, 39, 47, 55, 63, 
                         6, 14, 22, 30, 38, 46, 54, 62,
                         5, 13, 21, 29, 37, 45, 53, 61,
                         4, 12, 20, 28, 36, 44, 52, 60,
                         3, 11, 19, 27, 35, 43, 51, 59,
                         2, 10, 18, 26, 34, 42, 50, 58,
                         1,  9, 17, 25, 33, 41, 49, 57,
                         0,  8, 16, 24, 32, 40, 48, 56},

  rotateRight8x8[64] = {56, 48, 40, 32, 24, 16,  8,  0, 
                        57, 49, 41, 33, 25, 17,  9,  1,
                        58, 50, 42, 34, 26, 18, 10,  2,
                        59, 51, 43, 35, 27, 19, 11,  3,
                        60, 52, 44, 36, 28, 20, 12,  4,
                        61, 53, 45, 37, 29, 21, 13,  5,
                        62, 54, 46, 38, 30, 22, 14,  6,
                        63, 55, 47, 39, 31, 23, 15,  7},

  mirrorVert8x8[64]  = {56, 57, 58, 59, 60, 61, 62, 63, 
                        48, 49, 50, 51, 52, 53, 54, 55,
                        40, 41, 42, 43, 44, 45, 46, 47,
                        32, 33, 34, 35, 36, 37, 38, 39,
                        24, 25, 26, 27, 28, 29, 30, 31,
                        16, 17, 18, 19, 20, 21, 22, 23,
                         8,  9, 10, 11, 12, 13, 14, 15,
                         0,  1,  2,  3,  4,  5,  6,  7},

  mirrorHori8x8[64]  = { 7,  6,  5,  4,  3,  2,  1,  0, 
                        15, 14, 13, 12, 11, 10,  9,  8,
                        23, 22, 21, 20, 19, 18, 17, 16,
                        31, 30, 29, 28, 27, 26, 25, 24,
                        39, 38, 37, 36, 35, 34, 33, 32,
                        47, 46, 45, 44, 43, 42, 41, 40,
                        55, 54, 53, 52, 51, 50, 49, 48,
                        63, 62, 61, 60, 59, 58, 57, 56},

  mirrorDiag8x8[64]  = { 0,  8, 16, 24, 32, 40, 48, 56, 
                         1,  9, 17, 25, 33, 41, 49, 57,
                         2, 10, 18, 26, 34, 42, 50, 58,
                         3, 11, 19, 27, 35, 43, 51, 59,
                         4, 12, 20, 28, 36, 44, 52, 60,
                         5, 13, 21, 29, 37, 45, 53, 61,
                         6, 14, 22, 30, 38, 46, 54, 62,
                         7, 15, 23, 31, 39, 47, 55, 63},

  rotate180_8x8[64]  = {63, 62, 61, 60, 59, 58, 57, 56, 
                        55, 54, 53, 52, 51, 50, 49, 48,
                        47, 46, 45, 44, 43, 42, 41, 40,
                        39, 38, 37, 36, 35, 34, 33, 32,
                        31, 30, 29, 28, 27, 26, 25, 24,
                        23, 22, 21, 20, 19, 18, 17, 16,
                        15, 14, 13, 12, 11, 10,  9,  8,
                         7,  6,  5,  4,  3,  2,  1,  0};

u_char zig4x8[2*32] =   
{
   0,   1,   4,  10,
   2,   5,   9,  17,
   3,   8,  11,  18,
   6,  12,  16,  24,
   7,  15,  19,  25,
  13,  20,  23,  26,
  14,  22,  27,  30,
  21,  28,  29,  31,    /* End of zigzag part */
   0,   1,   4,   8,
   2,   5,  12,  16,
   9,   6,   3,  10,
  13,  20,  24,  17,
  14,   7,  11,  18,
  21,  28,  25,  22,
  15,  19,  23,  26,
  29,  30,  27,  31     /* End of inverse zigzag part */
};

u_char zig4x4[2*16] =   
{
   0,   1,   5,   6,
   2,   4,   7,  12,
   3,   8,  11,  13,
   9,  10,  14,  15,    /* End of zigzag part */
   0,   1,   4,   8,
   5,   2,   3,   6,
   9,  12,  13,  10,
   7,  11,  14,  15     /* End of inverse zigzag part */
};

u_char zig4x4Hori[2*16] =   
{
   0,   4,   8,  12,
   1,   5,   9,  13,
   2,   6,  10,  14,
   3,   7,  11,  15,    /* End of zigzag part */
   0,   4,   8,  12,
   1,   5,   9,  13,
   2,   6,  10,  14,
   3,   7,  11,  15     /* End of inverse zigzag part */
};

u_char zig4x4Vert[2*16] =   
{
   0,   1,   2,   3,
   4,   5,   6,   7,
   8,   9,  10,  11,
  12,  13,  14,  15,    /* End of zigzag part */
   0,   1,   2,   3,
   4,   5,   6,   7,
   8,   9,  10,  11,
  12,  13,  14,  15     /* End of inverse zigzag part */
};

u_char zig8x8[2*64] =   
{
   0,   1,   5,   6,  14,  15,  27,  28,
   2,   4,   7,  13,  16,  26,  29,  42,
   3,   8,  12,  17,  25,  30,  41,  43,
   9,  11,  18,  24,  31,  40,  44,  53,
  10,  19,  23,  32,  39,  45,  52,  54,
  20,  22,  33,  38,  46,  51,  55,  60,
  21,  34,  37,  47,  50,  56,  59,  61,
  35,  36,  48,  49,  57,  58,  62,  63,    /* End of zigzag part */
   0,   1,   8,  16,   9,   2,   3,  10,
  17,  24,  32,  25,  18,  11,   4,   5,
  12,  19,  26,  33,  40,  48,  41,  34,
  27,  20,  13,   6,   7,  14,  21,  28,
  35,  42,  49,  56,  57,  50,  43,  36,
  29,  22,  15,  23,  30,  37,  44,  51,
  58,  59,  52,  45,  38,  31,  39,  46,
  53,  60,  61,  54,  47,  55,  62,  63     /* End of inverse zigzag part */
};

u_char zig8x8Hori[2*64] =   
{
   0,   8,  15,  23,  30,  37,  45,  53,
   1,   9,  16,  24,  31,  38,  46,  55,
   2,  10,  17,  25,  32,  39,  47,  57,
   3,  11,  18,  26,  33,  41,  48,  58,
   4,  12,  19,  27,  35,  42,  50,  59,
   5,  13,  21,  28,  36,  44,  52,  60,
   6,  14,  22,  34,  43,  49,  54,  62,
   7,  20,  29,  40,  51,  56,  61,  63,    /* End of zigzag part */
   0,   8,  16,  24,  32,  40,  48,  56,
   1,   9,  17,  25,  33,  41,  49,   2,
  10,  18,  26,  34,  57,  42,  50,   3,
  11,  19,  27,  35,  43,  58,   4,  12,
  20,  28,  51,  36,  44,   5,  13,  21,
  59,  29,  37,  52,  45,   6,  14,  22,
  30,  53,  38,  60,  46,   7,  54,  15,
  61,  23,  31,  39,  47,  62,  55,  63     /* End of inverse zigzag part */
};

u_char zig8x8Vert[2*64] =   
{
   0,   1,   2,   3,   4,   5,   6,   7,
   8,   9,  10,  11,  12,  13,  14,  20,
  15,  16,  17,  18,  19,  21,  22,  29,
  23,  24,  25,  26,  27,  28,  34,  40,
  30,  31,  32,  33,  35,  36,  43,  51,
  37,  38,  39,  41,  42,  44,  49,  56,
  45,  46,  47,  48,  50,  52,  54,  61,
  53,  55,  57,  58,  59,  60,  62,  63,    /* End of zigzag part */
   0,   1,   2,   3,   4,   5,   6,   7,
   8,   9,  10,  11,  12,  13,  14,  16,
  17,  18,  19,  20,  15,  21,  22,  24,
  25,  26,  27,  28,  29,  23,  32,  33,
  34,  35,  30,  36,  37,  40,  41,  42,
  31,  43,  44,  38,  45,  48,  49,  50,
  51,  46,  52,  39,  53,  56,  54,  57,
  47,  58,  59,  60,  61,  55,  62,  63     /* End of inverse zigzag part */
};

u_char zig4x4Intra[2*16] =   
{
    15,    0,    4,    5,
     1,    3,    6,   11,
     2,    7,   10,   12,
     8,    9,   13,   14,
     1,    4,    8,    5,
     2,    3,    6,    9,
    12,   13,   10,    7,
    11,   14,   15,    0
};

u_char zig4x4IntraHori[2*16] =   
{
    15,    3,    7,   11,
     0,    4,    8,   12,
     1,    5,    9,   13,
     2,    6,   10,   14,
     4,    8,   12,    1,
     5,    9,   13,    2,
     6,   10,   14,    3,
     7,   11,   15,    0
};

u_char zig4x4IntraVert[2*16] =   
{
    15,    0,    1,    2,
     3,    4,    5,    6,
     7,    8,    9,   10,
    11,   12,   13,   14,
     1,    2,    3,    4,
     5,    6,    7,    8,
     9,   10,   11,   12,
    13,   14,   15,    0
};

u_char zig8x8Intra[2*64] =   
{
    63,    0,    4,    5,   13,   14,   26,   27,
     1,    3,    6,   12,   15,   25,   28,   41,
     2,    7,   11,   16,   24,   29,   40,   42,
     8,   10,   17,   23,   30,   39,   43,   52,
     9,   18,   22,   31,   38,   44,   51,   53,
    19,   21,   32,   37,   45,   50,   54,   59,
    20,   33,   36,   46,   49,   55,   58,   60,
    34,   35,   47,   48,   56,   57,   61,   62,
     1,    8,   16,    9,    2,    3,   10,   17,
    24,   32,   25,   18,   11,    4,    5,   12,
    19,   26,   33,   40,   48,   41,   34,   27,
    20,   13,    6,    7,   14,   21,   28,   35,
    42,   49,   56,   57,   50,   43,   36,   29,
    22,   15,   23,   30,   37,   44,   51,   58,
    59,   52,   45,   38,   31,   39,   46,   53,
    60,   61,   54,   47,   55,   62,   63,    0
};

u_char zig8x8IntraHori[2*64] =   
{
    63,    7,   14,   22,   29,   36,   44,   52,
     0,    8,   15,   23,   30,   37,   45,   54,
     1,    9,   16,   24,   31,   38,   46,   56,
     2,   10,   17,   25,   32,   40,   47,   57,
     3,   11,   18,   26,   34,   41,   49,   58,
     4,   12,   20,   27,   35,   43,   51,   59,
     5,   13,   21,   33,   42,   48,   53,   61,
     6,   19,   28,   39,   50,   55,   60,   62,
     8,   16,   24,   32,   40,   48,   56,    1,
     9,   17,   25,   33,   41,   49,    2,   10,
    18,   26,   34,   57,   42,   50,    3,   11,
    19,   27,   35,   43,   58,    4,   12,   20,
    28,   51,   36,   44,    5,   13,   21,   59,
    29,   37,   52,   45,    6,   14,   22,   30,
    53,   38,   60,   46,    7,   54,   15,   61,
    23,   31,   39,   47,   62,   55,   63,    0
};

u_char zig8x8IntraVert[2*64] =   
{
    63,    0,    1,    2,    3,    4,    5,    6,
     7,    8,    9,   10,   11,   12,   13,   19,
    14,   15,   16,   17,   18,   20,   21,   28,
    22,   23,   24,   25,   26,   27,   33,   39,
    29,   30,   31,   32,   34,   35,   42,   50,
    36,   37,   38,   40,   41,   43,   48,   55,
    44,   45,   46,   47,   49,   51,   53,   60,
    52,   54,   56,   57,   58,   59,   61,   62,
     1,    2,    3,    4,    5,    6,    7,    8,
     9,   10,   11,   12,   13,   14,   16,   17,
    18,   19,   20,   15,   21,   22,   24,   25,
    26,   27,   28,   29,   23,   32,   33,   34,
    35,   30,   36,   37,   40,   41,   42,   31,
    43,   44,   38,   45,   48,   49,   50,   51,
    46,   52,   39,   53,   56,   54,   57,   47,
    58,   59,   60,   61,   55,   62,   63,    0
};

/*
 * Table shows which klt basis function set has used for
 * given klt-method. The order of methods is same as
 * directions of edges. (now only 1 and 2 are used).
 */

u_char kltLookUpTable[10] = 
{ 
  0,  1,  2,  1,  0,  1,  2,  1,  0,  0
};


/*
 * Overhead order of MCC methods for 8x8 prediction classes.
 * So called 8x8 method selection table for MCC.
 */

u_char overheadOrder8x8[AMOUNT_OF_8x8_OVH*(MAX_OF_8x8_METH_OVH+1)] = {   
  /* 8x8 ovh ord LUMA LOW_QP */
  10,  1,42, 2, 7, 6, 3,52,56, 8, 4, 0, 0, /*  0 */
  12,  1,42,50,48,46,44, 8, 4,49,45,52,56, /*  1 */
   6, 44,52,42, 1, 8,45, 0, 0, 0, 0, 0, 0, /*  2 */
  10, 45,52,42, 1,44, 8,50,46,48,67, 0, 0, /*  3 */
  10,  1,44,45,42,52, 8,49, 4,66,48, 0, 0, /*  4 */
  11, 52, 8,42, 1, 4, 6, 7,44,50,48,46, 0, /*  5 */
   6, 42,44,49,52, 1, 8, 0, 0, 0, 0, 0, 0, /*  6 */
  12, 45,42,56,52, 4, 8,46,44, 3, 1,68,69, /*  7 */
  
  /* 8x8 ovh ord LUMA MED_LOW_QP */
  10,  1, 7, 6, 3, 2, 4, 8,42, 9, 5, 0, 0, /* 12 */
  12,  1,50,44,48,46,49,45,42, 8, 4,52,56, /* 13 */
   9, 44,42, 1, 8,50,45,52, 6,46, 0, 0, 0, /* 14 */
   3, 45, 1,50, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 15 */
   5,  1,44,45,49,66, 0, 0, 0, 0, 0, 0, 0, /* 16 */
  11, 52, 8,42, 1, 7, 6, 4,49,45,48,46, 0, /* 17 */
   6,  1,52,44,42,45,48, 0, 0, 0, 0, 0, 0, /* 18 */
  12, 45,42,56,52, 4, 8,46,44, 7, 1,68,69, /* 19 */
  
  /* 8x8 ovh ord LUMA MED_QP */
  11,  1, 7, 6, 3, 2,66,68,67,69, 8, 4, 0, /* 24 */
  12,  1,50,44,48,46,49,45,42, 8, 4,52,56, /* 25 */
   7, 44,42, 8, 7,52, 6,50, 0, 0, 0, 0, 0, /* 26 */
   3, 45, 1,46, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 27 */
   6, 44,46,48,45, 1, 4, 0, 0, 0, 0, 0, 0, /* 28 */
  12, 52, 8,42,50,44,46,48, 4,49, 7,45, 6, /* 29 */
   5,  1,44,45, 7, 6, 0, 0, 0, 0, 0, 0, 0, /* 30 */
  11,  1,45,46,44, 8, 4,68,69,42,52,56, 0, /* 31 */
  
  /* 8x8 ovh ord LUMA MED_HIGH_QP */
  11,  1, 7, 3, 6, 2,66,68,67,69, 8, 4, 0, /* 36 */
  10,  1,50,44,48,46,42, 8, 4,49,45, 0, 0, /* 37 */
   7, 44, 8,42,50,52, 4,45, 0, 0, 0, 0, 0, /* 38 */
   7, 45, 7, 1,46,50,49,42, 0, 0, 0, 0, 0, /* 39 */
   3, 44,45, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 40 */
  12, 52,50,44, 8, 7, 6,48,46, 4,49,45,42, /* 41 */
   3,  1,52,44, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 42 */
  12, 45, 1,44,46,48,50,69,68, 7, 2, 4, 8, /* 43 */
  
  /* 8x8 ovh ord CHROMA LOW_QP */
  11,  1, 2, 3, 7, 6, 5, 9,67,66,69,68, 0, /* 48 */
  12,  1,50,46,48,44,45,49, 8, 4,42,52,56, /* 49 */
   8, 46,44,42, 1,45,49,50, 8, 0, 0, 0, 0, /* 50 */
   2,  1,45, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 51 */
   3,  1,44,68, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 52 */
  12, 52, 8,50,44,42,45,49,68, 1,48,46, 4, /* 53 */
   2,  1,44, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 54 */
  11,  1,46,44,45,56,52, 4, 8,42,48,50, 0, /* 55 */
  
  /* 8x8 ovh ord CHROMA MED_LOW_QP */
  11,  1,67,66,69,68, 2, 7, 3, 6, 4, 8, 0, /* 60 */
   1,  1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 61 */
   1, 45, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 62 */
   1,  1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 63 */
   9, 48,44,49, 4, 7, 1,45,46,56, 0, 0, 0, /* 64 */
   8,  8,50,44,49,45,42,68,52, 0, 0, 0, 0, /* 65 */
   4, 52,46,50, 8, 0, 0, 0, 0, 0, 0, 0, 0, /* 66 */
  12, 42,46,44,56,52,68,69, 3,50, 9,48, 5  /* 67 */
};


/*
 * Table of used huffman code table for each 8x8 prediction classes.
 */

u_char overheadNumbers8x8[AMOUNT_OF_8x8_OVH] = {   
  68, 68, 68, 68, 68, 68, 68, 68, 
  68, 68, 68, 68, 68, 68, 68, 68, 
  68, 68, 68, 68, 68, 68, 68, 68, 
  68, 68, 68, 68, 68, 68, 68, 68, 
  68, 68, 68, 68, 68, 68, 68, 68, 
  68,  1,  1,  1, 68, 68, 68, 68 
};


/*
 * Overhead order of MCC methods for 4x4 prediction classes.
 * So called 4x4 method selection table for MCC.
 */

u_char overheadOrder4x4[AMOUNT_OF_4x4_OVH*(MAX_OF_4x4_METH_OVH+1)] = {   
  /* 4x4 ovh ord LUMA LOW_QP */
  11,  3, 0,33,39,37,35,34,38, 1, 4, 8, 0, /*  0 */
   0,  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /*  1 */
  
  /* 4x4 ovh ord LUMA MED_LOW_QP */
  11,  3, 0,33,39,37,35,34,38, 1, 4, 8, 0, /*  8 */
   0,  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /*  9 */

  /* 4x4 ovh ord LUMA MED_QP */
  11,  3, 0,33,39,37,35,38,34, 4, 8, 1, 0, /* 16 */
   0,  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 17 */
  
  /* 4x4 ovh ord LUMA MED_HIGH_QP */
  11,  3, 0,33,39,35,37, 1, 4, 8,34,38, 0, /* 24 */
   0,  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 25 */
  
  /* 4x4 ovh ord CHROMA LOW_QP */
  11,  3, 0, 1, 4, 8,33,37,35,39,34,38, 0, /* 32 */
  10,  3, 1,35,37,39,33,34,38, 8, 4, 0, 0, /* 39 */
  
  /* 4x4 ovh ord CHROMA MED_LOW_QP */
   9,  0, 3,37,35,33,34, 1,38,39, 0, 0, 0, /* 40 */
  10,  3, 1,34,38,35,33,37,39, 4, 8, 0, 0  /* 47 */
};


/*
 * Table of used huffman code table for each 4x4 prediction classes.
 */

u_char overheadNumbers4x4[AMOUNT_OF_4x4_OVH] = {   
  68,  0,
  68,  0,
  68,  0,
  68,  0,
  68, 68, 
  68, 68 
};
  
  
/*
 * Context dependent IP methods.
 */

u_char ipContextLists[49*(1+MAX_IP_LIST_SIZE)] =    
{
  5,  9, 1,13, 2, 5,-1, /*   0 */
  6,  9, 1,13, 5, 3, 7, /*   1 */
  5,  9, 5, 1,13, 4,-1, /*   2 */
  5,  9, 1, 2, 5, 3,-1, /*   3 */
  5,  9, 1,13, 2, 8,-1, /*   4 */
  5,  9, 1,13, 5, 2,-1, /*   5 */
  5,  9, 1, 2,13, 8,-1, /*   6 */
  6,  9, 1, 5,13, 8, 7, /*   7 */
  6,  1,13, 7, 9, 5, 8, /*   8 */
  5,  5, 1,13, 9, 7,-1, /*   9 */
  5,  1, 5, 9,13, 7,-1, /*  10 */
  6,  1, 7, 9,13, 8, 4, /*  11 */
  6,  1, 9,13, 5, 7, 8, /*  12 */
  6,  1, 9, 7, 8, 6,13, /*  13 */
  5,  1, 9, 5,13, 3,-1, /*  14 */
  6,  1, 5, 9, 6,13, 7, /*  15 */
  4,  5, 1,13, 9,-1,-1, /*  16 */
  6,  5, 1, 9, 4,13, 3, /*  17 */
  6,  1, 5, 9,13, 3, 6, /*  18 */
  4,  5, 1,13, 9,-1,-1, /*  19 */
  5,  5, 1, 9,13, 7,-1, /*  20 */
  4,  9, 1, 5, 2,-1,-1, /*  21 */
  5,  1, 9, 5, 7,13,-1, /*  22 */
  5,  5, 1, 9, 4,13,-1, /*  23 */
  3,  1, 2, 3,-1,-1,-1, /*  24 */
  4,  1, 9, 2, 7,-1,-1, /*  25 */
  3,  1, 9, 5,-1,-1,-1, /*  26 */
  3,  1, 9, 7,-1,-1,-1, /*  27 */
  5,  1, 9, 5, 8,13,-1, /*  28 */
  5,  1, 8, 5,13, 7,-1, /*  29 */
  5,  5, 1,13, 6, 4,-1, /*  30 */
  4,  1, 5, 4,13,-1,-1, /*  31 */
  6,  1, 9, 3, 5, 8,13, /*  32 */
  5,  1, 9, 5,13, 8,-1, /*  33 */
  4,  1, 9, 6, 7,-1,-1, /*  34 */
  6,  9, 1, 5,13, 2, 3, /*  35 */
  6,  1, 9, 5,13, 7, 6, /*  36 */
  4,  5, 1, 9,13,-1,-1, /*  37 */
  5,  1, 9, 5,13, 2,-1, /*  38 */
  6,  1, 9,13, 2, 5, 7, /*  39 */
  4,  1, 9, 5,13,-1,-1, /*  40 */
  5,  1, 9, 2, 5, 6,-1, /*  41 */
  6,  1, 9, 5, 8, 2,13, /*  42 */
  5,  1, 8, 5, 9, 6,-1, /*  43 */
  5,  5,13, 1, 9, 7,-1, /*  44 */
  5,  1, 5, 2,13, 3,-1, /*  45 */
  5,  1, 8, 5, 3, 2,-1, /*  46 */
  5,  1, 5, 9,13, 8,-1, /*  47 */
  5,  1, 5, 9, 8, 4,-1  /*  48 */
};


char motionTableX[AMOUNT_OF_INTRA_MOTION] = {
 -8,  -8,  -8,  -8,  -8,  -8,  -8,  -8,  -8,  -7,  -6,  -5,  -4,  -3,  -2,  -1,
  0,   1,   2,   3,   4,   5,   6,   7,  -8,  -7,  -6,  -5,  -4,  -3,  -2,  -1,
  0,   1,   2,   3,   4,   5,   6,   7,  -8,  -7,  -6,  -5,  -4,  -3,  -2,  -1,
  0,   1,   2,   3,   4,   5,   6,   7,  -8,  -7,  -6,  -5,  -4,  -3,  -2,  -1,
  0,   1,   2,   3,   4,   5,   6,   7,  -8,  -7,  -6,  -5,  -4,  -3,  -2,  -1,
  0,   1,   2,   3,   4,   5,   6,   7,  -8,  -7,  -6,  -5,  -4,  -3,  -2,  -1,
  0,   1,   2,   3,   4,   5,   6,   7,  -8,  -7,  -6,  -5,  -4,  -3,  -2,  -1,
  0,   1,   2,   3,   4,   5,   6,   7,  -8,  -7,  -6,  -5,  -4,  -3,  -2,  -1,
  0,   1,   2,   3,   4,   5,   6,   7,  -9, -10, -11, -12, -13, -14, -15,  -9,
-10, -11, -12, -13, -14, -15,  -9, -10, -11, -12, -13, -14, -15,  -9, -10, -11,
-12, -13, -14, -15,  -9, -10, -11, -12, -13, -14, -15,  -9, -10, -11, -12, -13,
-14, -15,  -9, -10, -11, -12, -13, -14, -15,  -9, -10, -11, -12, -13, -14, -15,
-15, -14, -13, -12, -11, -10,  -9,   8,   9,  10,  11,  12,  13,  14,  15,  16,
-15, -14, -13, -12, -11, -10,  -9,   8,   9,  10,  11,  12,  13,  14,  15,  16,
-15, -14, -13, -12, -11, -10,  -9,   8,   9,  10,  11,  12,  13,  14,  15,  16,
-15, -14, -13, -12, -11, -10,  -9,   8,   9,  10,  11,  12,  13,  14,  15,  16,
-15, -14, -13, -12, -11, -10,  -9,   8,   9,  10,  11,  12,  13,  14,  15,  16,
-15, -14, -13, -12, -11, -10,  -9,   8,   9,  10,  11,  12,  13,  14,  15,  16,
-15, -14, -13, -12, -11, -10,  -9,   8,   9,  10,  11,  12,  13,  14,  15,  16,
-15, -14, -13, -12, -11, -10,  -9,   8,   9,  10,  11,  12,  13,  14,  15,  16
};


char motionTableY[AMOUNT_OF_INTRA_MOTION] = {
 -7,  -6,  -5,  -4,  -3,  -2,  -1,   0,  -8,  -8,  -8,  -8,  -8,  -8,  -8,  -8,
 -8,  -8,  -8,  -8,  -8,  -8,  -8,  -8,  -9,  -9,  -9,  -9,  -9,  -9,  -9,  -9,
 -9,  -9,  -9,  -9,  -9,  -9,  -9,  -9, -10, -10, -10, -10, -10, -10, -10, -10,
-10, -10, -10, -10, -10, -10, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11,
-11, -11, -11, -11, -11, -11, -11, -11, -12, -12, -12, -12, -12, -12, -12, -12,
-12, -12, -12, -12, -12, -12, -12, -12, -13, -13, -13, -13, -13, -13, -13, -13,
-13, -13, -13, -13, -13, -13, -13, -13, -14, -14, -14, -14, -14, -14, -14, -14,
-14, -14, -14, -14, -14, -14, -14, -14, -15, -15, -15, -15, -15, -15, -15, -15,
-15, -15, -15, -15, -15, -15, -15, -15,  -7,  -7,  -7,  -7,  -7,  -7,  -7,  -6,
 -6,  -6,  -6,  -6,  -6,  -6,  -5,  -5,  -5,  -5,  -5,  -5,  -5,  -4,  -4,  -4,
 -4,  -4,  -4,  -4,  -3,  -3,  -3,  -3,  -3,  -3,  -3,  -2,  -2,  -2,  -2,  -2,
 -2,  -2,  -1,  -1,  -1,  -1,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,
 -8,  -8,  -8,  -8,  -8,  -8,  -8,  -8,  -8,  -8,  -8,  -8,  -8,  -8,  -8,  -8,
 -9,  -9,  -9,  -9,  -9,  -9,  -9,  -9,  -9,  -9,  -9,  -9,  -9,  -9,  -9,  -9,
-10, -10, -10, -10, -10, -10, -10, -10, -10, -10, -10, -10, -10, -10, -10, -10,
-11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11,
-12, -12, -12, -12, -12, -12, -12, -12, -12, -12, -12, -12, -12, -12, -12, -12,
-13, -13, -13, -13, -13, -13, -13, -13, -13, -13, -13, -13, -13, -13, -13, -13,
-14, -14, -14, -14, -14, -14, -14, -14, -14, -14, -14, -14, -14, -14, -14, -14,
-15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15
};


int16 kltBasisFunc4x4Diag1[8 * SIZE_4x4] = { 
    -7446,   -6782,   -5521,   -4008,
    -9684,   -9742,   -9275,   -8279,
    -8636,   -9522,   -9999,   -9844,
    -5792,   -6883,   -7981,   -8698,
     7671,   10629,   12751,   13614,
     -308,    2831,    6332,    8951,
    -7576,   -5912,   -3198,    -195,
   -10076,   -9854,   -8802,   -6844, 
     -404,    4676,   10975,   14787,
    -7807,   -8455,   -5973,   -1321,
     2767,   -2287,   -7236,   -9149,
    14108,   11894,    7053,    1858, 
     7486,    4611,   -4944,  -14020,
    -1301,    4939,   10104,    8797,
    -6371,  -10502,   -8075,   -1814,
    15263,    7174,   -3040,   -7530, 
   -14798,   -6925,     761,    3136,
   -14831,   -5817,    6419,   13209,
    -5880,   -3268,    5521,   15180,
     2762,     474,    -160,    3570,
    11777,   13550,    1455,  -12289,
   -14832,   -9062,   -1594,   -1865,
    -1560,    -117,     874,    2662,
    -9338,    -700,    8017,   11306,
    -7647,   -2425,     211,   -4786,
      -23,   -1934,    7098,   16607,
    12002,    5112,   -9806,  -14096,
   -11748,    1774,    7019,    -980,
     1206,    5922,    3038,   -4477,
    -7347,   -7203,   -2352,    -338,
    17736,    9096,    3817,    6481,
     4302,   -1105,  -12065,  -16649 
};

int16 kltBasisFunc4x4Diag2[8 * SIZE_4x4] = { 
    -9234,   -9510,   -7990,   -5254,
    -8003,   -9860,   -9856,   -8025,
    -5686,   -8261,   -9830,   -9609,
    -3088,   -5600,   -8038,   -9413,
     1138,   -5589,  -10863,  -11885,
     7294,    1268,   -6328,  -10835,
    11539,    8172,    1518,   -5579,
    12083,   11409,    7617,    1426,
    -7056,   -1868,    9470,   14748,
    -1597,   -8429,   -1824,    9405,
    10264,    -599,   -8410,   -2847,
    14317,   10163,    -871,   -7365,
      835,    9063,    -132,  -15550,
    -9023,     266,    9533,     694,
      426,  -10131,    -809,    9097,
    16800,     329,  -10316,   -1071,
   -18954,   -2689,    3749,  -12140,
    -3475,  -10202,    4530,    6038,
     3412,    3871,     -94,    9498,
   -10296,    5539,    9254,    5779,
     6818,    9908,    5138,  -11113,
     9472,     250,    3678,    4220,
     5105,    4119,   -9920,   -4938,
   -11109,    3746,   -2091,  -18717,
    -1428,  -13192,   -9061,    1432,
    14056,     632,     578,    7189,
     9125,     780,     623,   14463,
    -1327,   -6568,  -14820,   -2341,
      756,   -7078,    7288,   -9916,
     9565,   -1489,   -4249,   13806,
   -12813,    3952,   -2182,   -5483,
    12110,  -12309,    7440,    1011
};

int16 kltBasisFunc8x8Diag1[16 * SIZE_8x8] = { 
    -3988,   -4922,   -5553,   -5857,
    -6051,   -6243,   -6425,   -6392,
    -2112,   -3129,   -4125,   -4897,
    -5484,   -5886,   -6080,   -6172,
     -230,    -986,   -1866,   -2736,
    -3563,   -4282,   -4889,   -5256,
     1877,    1264,     481,    -362,
    -1223,   -1954,   -2730,   -3446,
     3588,    3332,    2866,    2193,
     1313,     464,    -391,   -1248,
     4479,    4484,    4334,    4032,
     3579,    2846,    1945,    1026,
     4781,    4979,    5044,    4859,
     4691,    4426,    3772,    2943,
     4549,    4783,    5014,    5084,
     5111,    4975,    4800,    4420,  
      654,    2036,    3439,    4487,
     5386,    6036,    6497,    6643,
    -2787,   -1941,    -616,     796,
     2345,    3594,    4471,    5153,
    -4377,   -4832,   -4650,   -3800,
    -2549,   -1115,     366,    1544,
    -2884,   -4316,   -5566,   -5965,
    -5905,   -5346,   -4239,   -2874,
      478,   -1039,   -2834,   -4442,
    -5714,   -6149,   -6129,   -5604,
     3604,    2672,    1262,    -377,
    -2250,   -3796,   -4996,   -5321,
     5195,    5042,    4509,    3536,
     2135,     598,   -1338,   -2900,
     5365,    5566,    5721,    5450,
     4936,    4010,    2787,    1345,  
    -2576,   -1207,     570,    2722,
     4844,    6566,    7756,    8423,
    -4211,   -4809,   -4699,   -3482,
    -1603,     495,    2603,    4531,
     -974,   -2525,   -4275,   -5242,
    -5664,   -5085,   -3584,   -1739,
     3028,    2318,     901,    -875,
    -2896,   -4141,   -4836,   -4753,
     2908,    4471,    5045,    4527,
     3180,    1665,    -334,   -1972,
    -1457,     581,    2949,    4812,
     5852,    5785,    4840,    3283,
    -5321,   -4555,   -2706,    -322,
     2227,    4080,    5265,    5427,
    -6539,   -6751,   -6317,   -5167,
    -3426,   -1345,    1092,    2799,  
     3083,    3465,    2654,     453,
    -2732,   -5606,   -7631,   -8772,
     -132,    2363,    4754,    5629,
     4866,    3050,     513,   -2415,
    -3503,   -3410,   -1544,    1081,
     3843,    5398,    5687,    4285,
     -515,   -2570,   -4813,   -5111,
    -3657,   -1671,     964,    2592,
     3072,    3302,    1462,   -1612,
    -4490,   -5553,   -4702,   -2989,
      642,    3231,    5556,    5633,
     3509,     335,   -2681,   -3834,
    -4569,   -2718,     559,    3655,
     6023,    6184,    4172,    1430,
    -8259,   -7756,   -6149,   -3755,
     -692,    1960,    4024,    4255,  
    -5377,   -4591,   -3777,   -2454,
     -782,    1417,    3478,    4406,
    -5794,   -4378,   -3263,   -2427,
    -1920,   -1151,     176,    1756,
    -7267,   -4906,   -1644,    1014,
     2419,    2678,    2141,    2215,
    -8930,   -7555,   -4717,    -836,
     3431,    6320,    7463,    6997,
    -5998,   -6317,   -5757,   -3655,
     -613,    2924,    6994,    9282,
    -1553,   -1066,   -1159,    -972,
     -534,     185,    2042,    4689,
    -1370,     274,    1824,    3127,
     3667,    3943,    3554,    3231,
    -3706,   -2350,    -821,    1142,
     3127,    4617,    5342,    5075,  
     1076,    -968,   -2881,   -2943,
     -447,    2998,    6964,    9880,
     5817,    4699,    1217,   -2993,
    -5883,   -6348,   -4790,   -1065,
     4016,    5275,    6300,    5121,
     1887,   -2125,   -5845,   -7056,
     1336,     -42,      91,    1631,
     3536,    4068,    2171,    -667,
     4280,    2253,   -1369,   -3837,
    -4797,   -3421,    -643,    1017,
     4105,    5300,    4451,    1565,
    -2677,   -6039,   -7390,   -6249,
    -1947,      -1,    2687,    4116,
     3861,    1650,   -2639,   -6084,
    -6631,   -6556,   -4469,   -1907,
      944,    2490,    2277,     444,  
    -9786,   -9086,   -5449,   -2010,
     -365,     249,     -43,    -851,
    -6767,   -7274,   -5749,   -2738,
      697,    2947,    4004,    3570,
     -225,   -1645,   -1783,    -769,
     1322,    4021,    6915,    7875,
     6005,    3365,    1000,      16,
      485,    2081,    4926,    7270,
     8092,    7206,    4530,    1584,
     -395,   -1081,      80,    2015,
     4475,    4674,    4112,    2283,
     -586,   -3031,   -3991,   -3428,
     1296,     970,     576,       5,
    -1436,   -3818,   -5922,   -6722,
      150,    -140,    -694,   -1111,
    -1768,   -2916,   -4552,   -6169,  
     7360,    8742,    5987,    2424,
     -579,    -991,    1037,    3463,
    -3355,     519,    2949,    2461,
     -961,   -4457,   -6336,   -5408,
    -8797,   -8645,   -5226,   -1435,
      672,    -342,   -3882,   -6764,
     -705,   -4333,   -5554,   -3965,
     -215,    3239,    4097,    2168,
     7093,    5343,    2572,     759,
      293,    2134,    5662,    7957,
     4518,    4345,    3505,    2678,
     1509,     756,    1411,    3792,
     2886,    1098,    -760,   -1530,
    -2145,   -2687,   -2744,   -1561,
     3902,    2261,    -316,   -2398,
    -4206,   -5220,   -5562,   -5414,  
     -926,   -2819,   -4908,   -5669,
    -4002,     124,    4788,    7983,
     1127,    1922,    2107,     247,
    -2461,   -3761,   -2876,     -83,
     -722,    -650,    1359,    3732,
     5096,    4321,    2024,     -40,
     2268,     268,   -2764,   -3805,
    -1949,    1083,    3592,    3879,
       40,    3020,    3488,     235,
    -3807,   -6033,   -5141,   -2380,
    -7468,   -5884,    -632,    3683,
     4965,    1967,   -2202,   -4186,
     -780,   -5779,   -8078,   -5696,
     -248,    4004,    5224,    3749,
    12165,    8941,    2547,   -2395,
    -4725,   -2651,    1828,    4538,  
     1477,    5100,    7761,    8270,
     5729,    1433,   -2194,   -3637,
    -7121,   -6169,   -3530,    -573,
      594,    -860,   -3433,   -5148,
      917,   -1642,   -3379,   -2812,
    -1537,    -982,   -2044,   -3284,
     6903,    6853,    4822,    3162,
     2985,    4290,    4621,    3704,
     -934,    1986,    3187,    1876,
      273,     195,    1688,    3749,
    -6445,   -5701,   -2652,   -1129,
    -2463,   -5465,   -7199,   -6563,
    -3365,   -4123,   -3277,    -684,
     2227,    2301,    -977,   -4762,
     2087,    1240,     642,    1438,
     3273,    6253,    8260,    6553,  
    -2180,   -3131,    -551,    3613,
     6407,    3459,   -4060,  -11296,
     4520,    3033,   -1809,   -3532,
     -435,    4622,    7885,    5070,
    -1241,    2449,    4386,    1762,
    -3152,   -6138,   -4797,    -513,
    -4720,   -5919,   -3354,    1188,
     4457,    3159,   -1917,   -4795,
     3821,    1776,   -2345,   -4279,
    -2749,    1378,    5241,    4748,
      -61,    2890,    5240,    4071,
      470,   -2454,   -2235,    1028,
    -2632,   -5354,   -4256,    -306,
     3269,    3521,      56,   -2410,
    10713,    5324,   -2223,   -5782,
    -5240,   -1495,    1889,    1811,  
     3477,    -229,   -5661,   -8213,
    -6452,   -3132,     384,    2678,
     3594,    5444,    5076,    2911,
     1392,    2303,    4499,    6802,
    -4411,   -4321,   -2558,   -1453,
    -1266,   -1897,   -1292,    1648,
     2130,    -311,   -2600,   -2998,
    -1860,   -1635,   -3532,   -4574,
     3177,    5425,    4545,    3060,
     3225,    4898,    5571,    3457,
    -7056,   -2844,     359,     178,
    -1491,   -2151,     -47,    2766,
    -7143,   -5578,   -2595,    -673,
     -833,   -3613,   -7607,   -8116,
    -1598,    -222,    2582,    5470,
     7703,    8057,    5143,     380,  
    -8504,   -3442,    1280,    3810,
     4776,    3274,     872,   -1723,
    -5709,    -546,    3564,    4448,
     2597,     711,   -1100,   -2687,
    -6604,   -1432,    4691,    8911,
     8296,    3223,   -2850,   -6711,
    -6206,   -3124,      78,    3104,
     5100,    3312,   -1880,   -7444,
    -3772,     411,    4855,    5889,
     3979,    1180,   -2279,   -6167,
    -4469,   -2250,    2031,    5985,
     6813,    4325,     190,   -4038,
     -927,     643,     931,    1435,
     1326,     223,   -2227,   -5452,
    -2731,     294,    3644,    4246,
     3053,     823,   -2876,   -6649,  
     9954,    4500,   -1551,   -2745,
    -1298,    -403,     -69,     127,
     4439,    2228,   -2582,   -5796,
    -4271,    -642,    2202,    3427,
    -1353,      74,   -1662,   -5025,
    -6330,   -3864,    1661,    5203,
    -5784,    -882,    3227,    3239,
      730,    -282,    3001,    8095,
    -9040,   -4729,    3688,    6813,
     4083,    1054,   -1111,      95,
    -3249,   -2460,    2557,    7671,
     7689,    2387,   -3883,   -7254,
     2490,    2659,    2918,    4393,
     5409,    4221,    -980,   -7380,
      398,     907,     114,   -1169,
    -1859,   -1960,   -3025,   -6102,  
    -1911,   -1529,     424,    1953,
      618,    -911,   -1382,   -1162,
      513,   -4261,   -6670,   -6310,
    -4583,   -5349,   -8379,   -9999,
     1055,    1975,   -1068,   -3876,
    -2456,     982,    3146,    1562,
    -8271,   -6107,   -3174,   -2581,
    -4647,   -6029,   -2870,    1213,
    -1139,   -2789,   -3679,   -2951,
    -1155,   -1665,   -4162,   -5477,
    -3117,     543,    2402,     519,
    -1455,    -504,    2277,    2211,
   -10745,   -9553,   -4243,   -1146,
    -2432,   -5771,   -7356,   -5485,
     1893,    -975,   -2403,    -522,
     2066,    1989,   -1222,   -4995,  
     2115,   -2296,   -2955,    2009,
     8001,    6730,    -939,   -9805,
     5654,    4378,   -2186,   -7467,
    -7160,   -2119,    1747,    -130,
    -1631,   -1517,     479,    1027,
     -931,   -2501,   -2399,    -975,
     3258,   -1211,   -6475,   -6759,
    -2136,    2069,    3637,    4206,
     1543,    3836,    3085,    -632,
    -4887,   -5327,   -1583,    2940,
    -3515,   -4624,   -3621,    -348,
     2788,    2748,      66,    -762,
      973,    2713,    -887,   -5226,
    -5599,   -3018,    1041,    3740,
   -11443,   -2424,    7235,    7819,
     2920,   -1939,   -2607,    -207  
};

int16 kltBasisFunc8x8Diag2[16 * SIZE_8x8] = { 
     -258,    1540,    3133,    4404,
     5305,    5895,    6083,    5860,
    -1855,     -29,    1770,    3620,
     4862,    5607,    6114,    6150,
    -3595,   -2031,     -36,    2032,
     3840,    5133,    5884,    6165,
    -4610,   -3802,   -2166,     -48,
     1880,    3778,    5061,    5627,
    -5047,   -4806,   -3954,   -2337,
     -250,    1731,    3482,    4693,
    -5215,   -5265,   -4970,   -4003,
    -2370,    -399,    1649,    3313,
    -5273,   -5388,   -5391,   -4941,
    -3845,   -2217,    -348,    1511,
    -4996,   -5297,   -5544,   -5332,
    -4851,   -3524,   -2101,    -562,  
     2947,    2580,     881,   -1222,
    -3301,   -5794,   -7101,   -6674,
     2428,    4216,    3575,    1280,
    -1316,   -3584,   -6105,   -7066,
      244,    3242,    5222,    4408,
     1870,   -1032,   -3607,   -5625,
    -1939,     842,    3792,    5671,
     5050,    2289,    -645,   -2860,
    -4042,   -2092,    1058,    4015,
     6039,    5310,    2866,      33,
    -5371,   -4432,   -1902,    1228,
     4101,    6053,    5424,    2948,
    -6023,   -5742,   -4416,   -1849,
     1174,    3947,    5736,    4838,
    -5795,   -5990,   -5412,   -3948,
    -1807,    1047,    3707,    5085,  
    -1932,   -3381,   -3601,   -2788,
     -115,    4689,    7654,    7719,
      774,   -1435,   -3647,   -4148,
    -3003,    -213,    4835,    7887,
     2551,    2210,    -847,   -3870,
    -4625,   -3706,    -476,    4377,
     1678,    4290,    3576,    -122,
    -3327,   -4472,   -3678,    -784,
    -1445,    2779,    5707,    4774,
      829,   -3047,   -4412,   -3535,
    -5553,   -1786,    3459,    6320,
     5605,    1143,   -2650,   -3650,
    -7463,   -5823,   -1792,    3192,
     6125,    5280,    1637,   -1577,
    -7116,   -7024,   -5179,   -1500,
     2643,    5246,    4650,    1992,  
     6033,    3883,     368,   -3539,
    -3155,    2411,    7788,    9131,
     4184,    6306,    3823,   -1065,
    -5269,   -4354,    1998,    7424,
      186,    3320,    6236,    2903,
    -2650,   -6512,   -4817,    1291,
    -2285,   -1844,    2043,    5405,
     2037,   -3251,   -6381,   -4864,
    -1302,   -4444,   -3623,     842,
     4472,    1095,   -3514,   -6131,
     2798,   -2117,   -5619,   -4546,
     -355,    3426,     386,   -3962,
     5688,    2650,   -2388,   -5434,
    -4733,    -935,    1691,    -745,
     6089,    5178,    1778,   -2382,
    -4859,   -4222,   -1484,     799,  
    -7229,   -7026,   -6590,   -5245,
    -2637,    -319,     995,    1683,
    -7035,   -6080,   -5361,   -4840,
    -3404,   -1414,     794,    2086,
    -6594,   -5599,   -3620,   -3026,
    -2240,   -1110,     498,    2236,
    -4711,   -4979,   -2759,   -1036,
     -151,     506,    1321,    2262,
    -1919,   -3469,   -2799,      -4,
     2118,    2904,    3384,    3621,
     1566,   -1304,   -2442,    -946,
     2562,    5065,    6037,    5400,
     5097,    1984,    -830,   -1074,
      957,    4983,    8009,    7927,
     5340,    5030,    2255,     137,
      337,    2691,    6978,    9602,  
     1293,   -2794,   -2672,    1871,
     4124,     287,   -5095,   -8047,
     5257,    1071,   -4185,   -3512,
     2905,    5901,    1454,   -4765,
     4036,    5884,    1222,   -4723,
    -3737,    3384,    7131,    2531,
     -597,    3393,    6072,    1037,
    -5234,   -4045,    3255,    6270,
    -2860,   -3386,    2252,    5400,
      600,   -5205,   -3697,    2098,
      912,   -5563,   -4791,    1090,
     4419,     553,   -4469,   -3151,
     6346,    -260,   -6603,   -5455,
     -262,    3281,     272,   -3534,
     8004,    5352,   -2053,   -6639,
    -5455,    -862,    2053,     393,  
     -868,    4144,    8522,    6968,
     3691,    1340,    -334,   -1409,
    -6275,   -1597,    4759,    7530,
     4992,    2073,     -63,   -1539,
    -7271,   -5159,    -793,    3616,
     3399,    1047,    -599,   -2240,
    -6133,   -5147,   -2739,    -310,
      545,   -1332,   -3319,   -4086,
    -4636,   -4390,   -1492,     561,
       32,   -2596,   -5818,   -6908,
    -1689,   -3173,   -1581,    2288,
     3697,     941,   -5031,   -8602,
     2340,     468,   -1024,    2179,
     5855,    6082,    1529,   -6048,
     2900,    3270,    1506,    1114,
     5146,    7500,    7336,    2093,  
     4625,   -1191,   -3797,    -744,
     4061,    2553,   -3944,   -7857,
     1337,    4819,   -1280,   -4979,
     -718,    4975,    4419,   -1818,
    -5008,     454,    5551,    -899,
    -5271,   -1176,    5594,    5452,
    -5259,   -5687,     -70,    5806,
     -486,   -5024,   -1243,    4915,
      157,   -3283,   -5538,    -567,
     5479,    -389,   -4150,   -1332,
     1597,    4654,    -422,   -4474,
    -1078,    4210,    -715,   -3445,
    -5597,    3663,    8066,    2483,
    -2416,    -866,    2541,    -803,
    -9689,   -4519,    5811,    9064,
     3686,   -1006,    -559,    1581,  
     -128,   -2029,   -4609,   -4509,
    -2954,    -655,     599,    1535,
     3123,    1674,     -56,   -1770,
    -3243,   -2879,    -543,    1474,
    -3091,    4115,    2742,    1993,
     1248,     109,    -927,     787,
   -10348,   -1107,    4674,    2509,
     2520,    3728,    3320,    1593,
   -10861,   -6741,    1762,    3858,
     -105,    1253,    4856,    5410,
    -4743,   -5861,    -163,    3882,
     1077,   -4467,   -1285,    4246,
     1700,    -315,    1506,    5439,
     5113,   -2022,   -8971,   -3967,
     3615,    2794,    3176,    6422,
     7745,    3980,   -4040,   -9670,  
     7781,    7338,   -1087,   -7308,
    -3909,     970,   -2704,   -7447,
      -86,    6461,    5618,   -2843,
    -7042,   -1091,    4200,     532,
    -1515,   -4236,    1503,    2909,
    -2278,   -3154,    3309,    7899,
     2998,   -2500,   -7458,   -2031,
      669,    -882,     -97,    5765,
     2099,    3811,   -1376,   -7254,
    -3691,    -665,    -247,    1815,
    -2424,   -1031,    3365,      58,
    -5011,   -3131,   -1339,      92,
     1585,   -6151,   -4444,    2901,
     3694,    -666,    -535,     230,
     8274,    -155,   -7845,   -3897,
     4677,    6335,    2221,    2367,  
    -8876,   -4230,      84,     -79,
     2514,    6103,   -1708,  -10442,
    -6229,   -5525,     252,     718,
     -808,    3403,    7097,     238,
     -400,   -2432,     -72,    4181,
      967,   -3365,    1133,    5966,
     1863,    2173,    1521,    4471,
     7525,     363,   -6202,   -2691,
     -235,    3159,    3367,    3069,
     5971,    7558,   -1004,   -7811,
    -1963,     -13,    3333,    3024,
     1414,    3659,    5198,   -1582,
     1980,   -2315,     -46,    1635,
     -116,   -3430,   -2156,     972,
     6389,     692,   -3177,   -1765,
    -1610,   -4405,   -8623,   -6952,  
     4171,    2325,    4273,    6648,
     7084,    6191,     797,   -3897,
     6649,     713,   -2327,    -333,
     2272,    3153,    2561,    -825,
     5145,    3656,   -3487,   -5416,
    -4067,   -4002,   -2818,   -2207,
    -1381,    2927,    -790,   -5739,
    -3966,   -3634,   -5645,   -6373,
    -7143,   -2366,     409,   -3569,
    -4235,     958,    1344,   -3227,
    -5410,   -5814,    -119,    -439,
    -4658,   -1044,    6806,    6552,
      536,   -2511,     812,    3239,
     -613,   -3754,    2418,   10719,
     3655,    1895,    2426,    5985,
     4742,     712,   -1167,    5226,  
    -2958,    5865,    6134,   -3475,
    -5558,    2275,    -741,  -10666,
     2009,   -4104,    4822,    4273,
    -5519,   -4250,    5912,    3051,
     2207,    1971,   -5448,    1686,
     1131,   -4645,    -665,    8712,
     -723,    1804,    2397,   -6522,
     -523,     -21,   -3063,    1079,
      -56,   -4059,     758,    3299,
    -5991,    -580,    1477,    1142,
     4190,    -516,   -5427,     630,
     4101,   -3796,     681,    3728,
    -1169,    6868,    -114,   -5001,
      911,    4392,   -2716,    2039,
   -10622,    1256,    6759,    -928,
    -4291,     336,    4419,   -2580,  
     1854,   -4048,   -8512,   -1547,
     6651,    9337,    3675,   -4151,
     5847,    1479,   -3354,   -3858,
     1345,    4015,    3688,   -2041,
      875,    2505,     695,    1105,
     2303,    1462,   -2938,   -6438,
      561,   -3803,   -1293,    -611,
     4150,    5554,    -302,   -9618,
     7104,   -2263,   -5150,   -1973,
    -1558,    4106,    5238,   -2735,
     8010,    2274,   -3544,   -2642,
      -95,   -3061,    1826,    1440,
     1941,    1161,   -3526,   -1637,
     3863,    3994,   -3355,   -3407,
    -3451,   -1442,   -3976,   -4705,
     2353,    8035,    5728,   -4700,  
    -5118,   -6090,   -4115,   -4382,
    -5205,   -2483,     -59,   -2058,
     4606,   -1623,    -876,    3231,
     2234,   -1162,    2032,    3174,
     9611,    5320,    -137,    3926,
     8245,    5614,    2512,    4004,
     5724,    4613,      14,    -783,
     5032,    6242,    2125,    1153,
    -2900,    -485,   -3177,   -4782,
    -1614,    1737,   -2359,   -6187,
    -5881,   -3908,   -3176,   -7167,
    -4616,      61,   -1092,   -8499,
    -1642,   -2208,    2426,    -412,
    -4067,     495,    3080,   -2244,
     2702,     261,    4117,    7028,
     3458,    1930,    6421,    5060,  
     7201,    3156,   -6569,   -6319,
     2970,    8406,    4660,   -2395,
    -6164,    4919,    1543,   -2118,
    -1181,    2661,    2610,   -3322,
       94,   -5931,     617,    1046,
     3365,    3188,    -434,   -5088,
     4125,    4664,   -3934,     385,
     1008,    4560,    3988,   -2153,
    -5041,    4235,    7388,    -514,
      882,    -781,    1579,    1351,
    -6031,   -4582,    2337,    6004,
      465,    1780,   -2359,   -2088,
      176,    4717,   -1311,   -1590,
      415,   -1261,    3495,   -1729,
    -6759,    8200,   11233,   -1115,
    -4370,   -2511,   -2002,    4005  
};
